#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A class for getting numeric or categorical ratings, e.g., a 1-to-7 scale."""

# Part of the PsychoPy library
# Copyright (C) 2015 Jonathan Peirce
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

import copy
import numpy as np

from psychopy import core, logging, event
from .basevisual import MinimalStim, ColorMixin, BaseVisualStim
from .rect import Rect
from .grating import GratingStim
from .elementarray import ElementArrayStim
from .circle import Circle
from .shape import ShapeStim
from .text import TextStim
from ..tools.attributetools import logAttrib, setAttribute, attributeSetter
from ..constants import FINISHED, STARTED, NOT_STARTED



class Slider(MinimalStim, ColorMixin):
    """A class for obtaining ratings, e.g., on a 1-to-7 or categorical scale.

    A simpler alternative to RatingScale, to be customised with code rather
    than with arguments.

    A RatingScale instance is a re-usable visual object having a ``draw()``
    method, with customizable appearance and response options. ``draw()``
    displays the rating scale, handles the subject's mouse or key responses,
    and updates the display. When the subject accepts a selection,
    ``.noResponse`` goes ``False`` (i.e., there is a response).

    You can call the ``getRating()`` method anytime to get a rating,
    ``getRT()`` to get the decision time, or ``getHistory()`` to obtain
    the entire set of (rating, RT) pairs.

    For other examples see Coder Demos -> stimuli -> ratingsNew.py.

    :Authors:
        - 2018: Jon Peirce
    """

    def __init__(self,
                 win,
                 ticks=(1, 2, 3, 4, 5),
                 labels=None,
                 pos=None,
                 size=None,
                 units=None,
                 flip=False,
                 style='rating', styleTweaks=[],
                 granularity=0,
                 readOnly=False,
                 color='White',
                 fillColor='Red',
                 borderColor='White',
                 colorSpace='rgb',
                 opacity=None,
                 font='Helvetica Bold',
                 depth=0,
                 name=None,
                 labelHeight=None,
                 labelWrapWidth=None,
                 autoDraw=False,
                 autoLog=True):
        """

        Parameters
        ----------
        win : psychopy.visual.Window
            Into which the scale will be rendered

        ticks : list or tuple
            A set of values for tick locations. If given a list of numbers then
            these determine the locations of the ticks (the first and last
            determine the endpoints and the rest are spaced according to
            their values between these endpoints.

        labels : a list or tuple
            The text to go with each tick (or spaced evenly across the ticks).
            If you give 3 labels but 5 tick locations then the end and middle
            ticks will be given labels. If the labels can't be distributed
            across the ticks then an error will be raised. If you want an
            uneven distribution you should include a list matching the length
            of ticks but with some values set to None

        pos : XY pair (tuple, array or list)

        size : w,h pair (tuple, array or list)
            The size for the scale defines the area taken up by the line and
            the ticks.
            This also controls whether the scale is horizontal or vertical.

        units : the units to interpret the pos and size

        flip : bool
            By default the labels will be below or left of the line. This
            puts them above (or right)

        granularity : int or float
            The smallest valid increments for the scale. 0 gives a continuous
            (e.g. "VAS") scale. 1 gives a traditional likert scale. Something
            like 0.1 gives a limited fine-grained scale.

        color :
            Color of the line/ticks/labels according to the color space

        font : font name

        autodraw :

        depth :

        name :

        autoLog :
        """
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        super(Slider, self).__init__(name=name, autoLog=False)

        self.win = win
        self.ticks = np.asarray(ticks)
        self.labels = labels

        if pos is None:
            self.__dict__['pos'] = (0, 0)
        else:
            self.__dict__['pos'] = pos

        if units is None:
            self.units = win.units
        else:
            self.units = units

        if size is None:
            defaultSizes = {
                None: [self.win.size[0]*1.0, self.win.size[1]*0.1],
                'height': [1.0, 0.1],
                'norm': [1.0, 0.1],
                'deg': [3.14, 0.314],
                'cm': [10, 1],
                'pix': [self.win.size[0]*1.0, self.win.size[1]*0.1]
            }
            self._size = defaultSizes[self.units]
        else:
            self._size = size

        self.flip = flip
        self.granularity = granularity
        self.colorSpace = colorSpace
        self.color = color
        self.fillColor = fillColor
        self.borderColor = borderColor
        self.opacity = opacity
        self.font = font
        self.autoDraw = autoDraw
        self.depth = depth
        self.name = name
        self.autoLog = autoLog
        self.readOnly = readOnly

        self.categorical = False  # will become True if no ticks set only labels
        self.rating = None  # current value (from a response)
        self.markerPos = None  # current value (maybe not from a response)
        self.rt = None
        self.history = []
        self.marker = None
        self.line = None
        self.tickLines = []
        self.tickLocs = None
        self.labelLocs = None
        self.labelWrapWidth = labelWrapWidth
        self.labelHeight = labelHeight or min(self._size)
        self._lineAspectRatio = 0.01
        self._updateMarkerPos = True
        self._dragging = False
        self.mouse = event.Mouse()
        self._mouseStateClick = None  # so we can rule out long click probs
        self._mouseStateXY = None  # so we can rule out long click probs

        self.validArea = None
        self._createElements()
        # some things must wait until elements created
        self.contrast = 1.0

        # set autoLog (now that params have been initialised)
        self.autoLog = autoLog
        if autoLog:
            logging.exp("Created %s = %s" % (self.name, repr(self)))
        self.status = NOT_STARTED
        self.responseClock = core.Clock()

        # set the style when everything else is set
        self.styleTweaks = []
        self.style = style
        self.styleTweaks += styleTweaks

    def __repr__(self, complete=False):
        return self.__str__(complete=complete)  # from MinimalStim

    @property
    def _lineL(self):
        """The length of the line (in the size units)
        """
        return max(self.size)

    @property
    def _tickL(self):
        """The length of the line (in the size units)
        """
        return min(self.size)

    @property
    def _lineW(self):
        """The length of the line (in the size units)
        """
        return max(self.size) * self._lineAspectRatio

    @property
    def horiz(self):
        """(readonly) determines from self.size whether the scale is horizontal"""
        return self.size[0] > self.size[1]

    @property
    def size(self):
        """The size for the scale defines the area taken up by the line and
            the ticks.
        """
        return self._size

    @property
    def opacity(self):
        BaseVisualStim.opacity.fget(self)
    @opacity.setter
    def opacity(self, value):
        BaseVisualStim.opacity.fset(self, value)
        self.fillColor = self._fillColor.copy()
        self.borderColor = self._borderColor.copy()
        self.foreColor = self._foreColor.copy()
    def setOpacity(self, value):
        self.opacity = value

    @property
    def foreColor(self):
        ColorMixin.foreColor.fget(self)
    @foreColor.setter
    def foreColor(self, value):
        ColorMixin.foreColor.fset(self, value)
        # Set color of each label
        if hasattr(self, 'labelObjs'):
            for lbl in self.labelObjs:
                lbl.color = self._foreColor.copy()

    @property
    def fillColor(self):
        ColorMixin.fillColor.fget(self)
    @fillColor.setter
    def fillColor(self, value):
        ColorMixin.fillColor.fset(self, value)
        # Set color of marker
        if hasattr(self, 'marker'):
            self.marker.fillColor = self._foreColor.copy()

    @property
    def borderColor(self):
        ColorMixin.borderColor.fget(self)
    @borderColor.setter
    def borderColor(self, value):
        ColorMixin.borderColor.fset(self, value)
        # Set color of lines
        if hasattr(self, 'line'):
            if self.style not in ["scrollbar"]: # Scrollbar doesn't have an outline
                self.line.color = self._borderColor.copy()
            self.line.fillColor = self._borderColor.copy()
            if self.style in ["slider", "scrollbar"]: # Slider and scrollbar need translucent fills
                self.line._fillColor.alpha *= 0.2
        if hasattr(self, 'tickLines'):
            self.tickLines.colors = self._borderColor.copy()
            self.tickLines.opacities = self._borderColor.alpha


    def reset(self):
        """Resets the slider to its starting state (so that it can be restarted
        on each trial with a new stimulus)
        """
        self.markerPos = None
        self.history = []
        self.rating = None
        self.rt = None
        self.responseClock.reset()
        self.status = NOT_STARTED

    def _createElements(self):
        if not self.tickLocs:
            self._setTickLocs()
        if self.horiz:
            lineSize = self._lineL, self._lineW
            tickSize = self._lineW, self._tickL
        else:
            lineSize = self._lineW, self._lineL
            tickSize = self._tickL, self._lineW
        self.line = GratingStim(win=self.win, pos=self.pos, color=self._borderColor.copy(), colorSpace=self.colorSpace,
                                size=lineSize, sf=0, units=self.units,
                                autoLog=False)
        self.tickLines = ElementArrayStim(win=self.win, units=self.units,
                                          nElements=len(self.ticks),
                                          xys=self.tickLocs,
                                          elementMask=None,
                                          colors=self._borderColor.copy(), colorSpace = self.colorSpace,
                                          opacities=self._borderColor.alpha,
                                          sizes=tickSize, sfs=0,
                                          autoLog=False)

        self.labelObjs = []
        if self.labels is not None:
            if not self.labelLocs:
                self._setLabelLocs()
            if self.horiz:
                alignHoriz = 'center'
                if not self.flip:
                    alignVert = 'top'
                    self.labelLocs -= [0, self._tickL]
                else:
                    alignVert = 'bottom'
                    self.labelLocs += [0, self._tickL]
            else:  # vertical
                alignVert = 'center'
                if not self.flip:
                    alignHoriz = 'right'
                    self.labelLocs -= [self._tickL, 0]
                else:
                    alignHoriz = 'left'
                    self.labelLocs += [self._tickL, 0]
            for tickN, label in enumerate(self.labels):
                if label is None:
                    continue

                obj = TextStim(self.win, label, font=self.font,
                               anchorHoriz=alignHoriz, anchorVert=alignVert,
                               units=self.units, color=self._foreColor.copy(), colorSpace=self.colorSpace,
                               pos=self.labelLocs[tickN, :],
                               height=self.labelHeight, 
                               wrapWidth=self.labelWrapWidth,
                               autoLog=False)
                self.labelObjs.append(obj)

        if self.units == 'norm':
            # convert to make marker round
            aspect = self.win.size[0] / self.win.size[1]
            markerSize = np.array([self._tickL, self._tickL * aspect])
        else:
            markerSize = self._tickL

        self.marker = Circle(self.win, units=self.units,
                             size=markerSize,
                             fillColor=self._fillColor,
                             autoLog=False)

        # create a rectangle to check for clicks
        self.validArea = Rect(self.win, units=self.units,
                              pos=self.pos,
                              size=[d * 1.1 for d in self.size],
                              lineColor='DarkGrey',
                              autoLog=False)

    @attributeSetter
    def pos(self, newPos):
        """Set position of slider

        Parameters
        ----------
        value: tuple, list
            The new position of slider
        """
        newPos = np.array(newPos)
        oldPos = self.__dict__['pos']
        self.__dict__['pos'] = newPos
        deltaPos = np.subtract(newPos, oldPos)
        self.line.pos += deltaPos
        self.validArea.pos += deltaPos
        self.marker.pos += deltaPos
        self.tickLines.xys += deltaPos
        for label in self.labelObjs:
            label.pos += deltaPos

    def _ratingToPos(self, rating):
        try:
            n = len(rating)
        except:
            n = 1
        pos = np.ones([n, 2], 'f') * self.pos

        scaleMag = self.ticks[-1] - self.ticks[0]
        scaleLow = self.ticks[0]
        if self.horiz:
            pos[:, 0] = (((rating - scaleLow) / scaleMag - 0.5) * self._lineL +
                         self.pos[0])
        else:
            pos[:, 1] = (((rating - scaleLow) / scaleMag - 0.5) * self._lineL +
                         self.pos[1])

        return pos

    def _posToRating(self, pos):
        scaleMag = self.ticks[-1] - self.ticks[0]
        scaleLow = self.ticks[0]
        if self.horiz:
            rating = (((pos[0] - self.pos[0]) / self._lineL + 0.5)
                      * scaleMag + scaleLow)
        else:
            rating = (((pos[1] - self.pos[1]) / self._lineL + 0.5)
                      * scaleMag + scaleLow)

        return rating

    def _setTickLocs(self):
        """ Calculates the locations of the line, tickLines and labels from
        the rating info
        """
        try:
            n = len(self.ticks)
        except TypeError:
            self.categorical = True
        if self.categorical:
            self.ticks = np.arange(len(self.labels))
            self.granularity = 1.0

        self.tickLocs = self._ratingToPos(self.ticks)

    def _setLabelLocs(self):
        """ Calculates the locations of the line, tickLines and labels from
        the rating info
        """
        if not self.labels:
            self.labelLocs = []
            return
        labelFractions = np.arange(len(self.labels)) / (len(self.labels) - 1)
        tickIndices = np.round(labelFractions * (len(self.tickLocs) - 1))
        self.labelLocs = self.tickLocs[tickIndices.astype('int')]

    def _granularRating(self, rating):
        """Handle granularity for the rating"""
        if rating is not None:
            if self.granularity > 0:
                rating = round(rating / self.granularity) * self.granularity
                rating = round(rating, 8)  # or gives 1.9000000000000001
            rating = max(rating, self.ticks[0])
            rating = min(rating, self.ticks[-1])
        return rating

    @attributeSetter
    def rating(self, rating):
        """The most recent rating from the participant or None.
        Note that the position of the marker can be set using current without
        looking like a change in the marker position"""
        rating = self._granularRating(rating)
        self.markerPos = rating
        if self.categorical and (rating is not None):
            rating = self.labels[int(round(rating))]
        self.__dict__['rating'] = rating

    @property
    def value(self):
        """Synonymous with .rating"""
        return self.rating
    @value.setter
    def value(self, val):
        self.rating = val

    @attributeSetter
    def markerPos(self, rating):
        """The position on the scale where the marker should be. Note that
        this does not alter the value of the reported rating, only its visible
        display.
        Also note that this position is in scale units, not in coordinates"""
        rating = self._granularRating(rating)
        if ('markerPos' not in self.__dict__ or not np.alltrue(
                self.__dict__['markerPos'] == rating)):
            self.__dict__['markerPos'] = rating
            self._updateMarkerPos = True

    def recordRating(self, rating, rt=None, log=None):
        """Sets the current rating value
        """
        rating = self._granularRating(rating)
        setAttribute(self, attrib='rating', value=rating, operation='', log=log)
        if rt is None:
            self.rt = self.responseClock.getTime()
        else:
            self.rt = rt
        self.history.append((rating, self.rt))
        self._updateMarkerPos = True

    def getRating(self):
        """Get the current value of rating (or None if no response yet)
        """
        return self.rating

    def getRT(self):
        """Get the RT for most recent rating (or None if no response yet)
        """
        return self.rt

    def draw(self):
        """Draw the Slider, with all its constituent elements on this frame
        """
        self.getMouseResponses()
        # self.validArea.draw()
        self.line.draw()
        self.tickLines.draw()
        if self.markerPos is not None:
            if self._updateMarkerPos:
                self.marker.pos = self._ratingToPos(self.markerPos)
                self._updateMarkerPos = False
            self.marker.draw()
        for label in self.labelObjs:
            label.draw()
        # we started drawing to reset clock on flip
        if self.status == NOT_STARTED:
            self.win.callOnFlip(self.responseClock.reset)
            self.status = STARTED

    def getHistory(self):
        """Return a list of the subject's history as (rating, time) tuples.

        The history can be retrieved at any time, allowing for continuous
        ratings to be obtained in real-time. Both numerical and categorical
        choices are stored automatically in the history.
        """
        return self.history

    def setReadOnly(self, value=True, log=None):
        """When the rating scale is read only no responses can be made and the
        scale contrast is reduced

        Parameters
        ----------
        value : bool (True)
            The value to which we should set the readOnly flag
        log : bool or None
            Force the autologging to occur or leave as default

        """
        setAttribute(self, 'readOnly', value, log)
        if value == True:
            self.contrast = 0.5
        else:
            self.contrast = 1.0

    @attributeSetter
    def contrast(self, contrast):
        """Set all elements of the Slider (labels, ticks, line) to a contrast

        Parameters
        ----------
        contrast
        """
        self.marker.contrast = contrast
        self.line.contrast = contrast
        self.tickLines.contrasts = contrast
        for label in self.labelObjs:
            label.contrast = contrast

    def getMouseResponses(self):
        """Instructs the rating scale to check for valid mouse responses.

        This is usually done during the draw() method but can be done by the
        user as well at any point in time. The rating will be returned but
        will ALSO automatically be set as the current rating response.

        While the mouse button is down we will alter self.markerPos
        but don't set a value for self.rating until button comes up

        Returns
        ----------
        A rating value or None
        """
        if self.readOnly:
            return
        click = bool(self.mouse.getPressed()[0])
        xy = self.mouse.getPos()

        if click:
            # Update current but don't set Rating (mouse is still down)
            # Dragging has to start inside a "valid" area (i.e., on the
            # slider), but may continue even if the mouse moves away from
            # the slider, as long as the mouse button is not released.
            if (self.validArea.contains(self.mouse, units=self.units) or
                    self._dragging):
                self.markerPos = self._posToRating(xy)  # updates marker
                self._dragging = True
        else:  # mouse is up - check if it *just* came up
            if self._dragging:
                self._dragging = False
                if self.markerPos is not None:
                    self.recordRating(self.markerPos)
                return self.markerPos
            else:
                # is up and was already up - move along
                return None

        self._mouseStateXY = xy

    # Overload color setters so they set sub-components
    @property
    def foreColor(self):
        ColorMixin.foreColor.fget(self)
    @foreColor.setter
    def foreColor(self, value):
        ColorMixin.foreColor.fset(self, value)
        # Set color for all labels
        if hasattr(self, "labelObjs"):
            for obj in self.labelObjs:
                obj.color = self._foreColor.copy()

    @property
    def fillColor(self):
        ColorMixin.fillColor.fget(self)
    @fillColor.setter
    def fillColor(self, value):
        ColorMixin.fillColor.fset(self, value)
        # Set color for marker
        if hasattr(self, "marker"):
            self.marker.fillColor = self._fillColor.copy()

    @property
    def borderColor(self):
        ColorMixin.borderColor.fget(self)
    @borderColor.setter
    def borderColor(self, value):
        ColorMixin.borderColor.fset(self, value)
        # Set color for lines
        if hasattr(self, "line"):
            self.line.color = self._borderColor.copy()
        if hasattr(self, "tickLines"):
            self.tickLines.colors = self._borderColor.copy()

    knownStyles = ['slider', 'rating', 'radio', 'scrollbar']
    legacyStyles = []
    knownStyleTweaks = ['labels45', 'triangleMarker']
    legacyStyleTweaks = ['whiteOnBlack']

    @attributeSetter
    def style(self, style):
        """Sets some predefined styles or use these to create your own.

        If you fancy creating and including your own styles that would be great!

        Parameters
        ----------
        style: string

            Known styles currently include:

                'rating': the marker is a circle
                'slider': looks more like an application slider control
                'whiteOnBlack': a sort of color-inverse rating scale
                'scrollbar': looks like a scrollbar for a window

            Styles cannot be combined in a list - they are discrete

        """
        self.__dict__['style'] = style

        # Legacy: If given a list (as was once the case), take the first style
        if isinstance(style, (list, tuple)):
            styles = style
            style = "rating"
            for val in styles:
                # If list contains a style, use it
                if val in self.knownStyles + self.legacyStyles:
                    style = val
                # Apply any tweaks
                if val in self.knownStyleTweaks + self.legacyStyleTweaks:
                    self.styleTweaks += val

        if style == 'rating':
            pass  # this is just the default

        if style == 'slider':
            # make it more like a slider using a box instead of line
            self.line = Rect(self.win, units=self.units,
                             pos=self.pos,
                             size=self.size,
                             fillColor=self._borderColor.copy(),
                             lineColor=None,
                             autoLog=False)
            self.line._fillColor.alpha *= 0.2
            if self.horiz:
                markerW = self.size[0] * 0.01
                markerH = self.size[1] * 0.8
            else:
                markerW = self.size[0] * 0.8
                markerH = self.size[1] * 0.01

            self.marker = Rect(self.win, units=self.units,
                               size=[markerW, markerH],
                               fillColor=self._fillColor,
                               lineColor=None,
                               autoLog=False)

        if style == 'radio':
            # no line, ticks are circles
            self.line.opacity = 0
            # ticks are circles
            self.tickLines.sizes = (self._tickL, self._tickL)
            self.tickLines.elementMask = 'circle'
            # marker must be smalle than a "tick" circle
            self.marker.size = self._tickL * 0.7
            self.marker.fillColor = self._fillColor.copy()

        if style == 'scrollbar':
            # Make marker the full height and 20% of the width of the slider
            markerSz = self.size[0]*0.2 if self.horiz else self.size[1]*0.2
            w, h = self.size
            self.marker = Rect(self.win, units=self.units,
                               size=[markerSz, h] if self.horiz else [w, markerSz],
                               fillColor=self._fillColor,
                               lineColor=None,
                               autoLog=False)
            # Make the line a translucent box
            self.line = Rect(self.win, units=self.units,
                             pos=self.pos,
                             size=[w+markerSz, h] if self.horiz else [w, h+markerSz],
                             fillColor=self._borderColor.copy(),
                             lineColor=None,
                             autoLog=False)
            self.line._fillColor.alpha *= 0.05
            self.tickLines = Rect(self.win, size=(0,0), lineColor=None, fillColor=None)

        # Legacy: If given a tweak, apply it as a tweak rather than a style
        if style in self.knownStyleTweaks + self.legacyStyleTweaks:
            self.styleTweaks.append(style)

    @attributeSetter
    def styleTweaks(self, styleTweaks):
        """Sets some predefined style tweaks or use these to create your own.

        If you fancy creating and including your own style tweaks that would be great!

        Parameters
        ----------
        styleTweaks: list of strings

            Known style tweaks currently include:

                'triangleMarker': the marker is a triangle
                'labels45': the text is rotated by 45 degrees

            Legacy style tweaks include:

                'whiteOnBlack': a sort of color-inverse rating scale

            Legacy style tweaks will work if set in code, but are not exposed in Builder as they are redundant

            Style tweaks can be combined in a list e.g. `['labels45']`

        """
        self.__dict__['styleTweaks'] = styleTweaks

        if 'triangleMarker' in styleTweaks:
            if self.horiz and self.flip:
                ori = -90
            elif self.horiz:
                ori = -90
            elif not self.horiz and self.flip:
                ori = 180
            else:
                ori = 0

            markerSize = min(self.size) * 2
            self.marker = ShapeStim(self.win, units=self.units,
                                    vertices=[[0, 0], [0.5, 0.5], [0.5, -0.5]],
                                    size=markerSize,
                                    ori=ori,
                                    fillColor=self._fillColor.copy(),
                                    lineColor=None,
                                    lineWidth=0,
                                    autoLog=False)

        if 'labels45' in styleTweaks:
            for label in self.labelObjs:
                if self.flip:
                    label.alignHoriz = 'left'
                else:
                    label.alignHoriz = 'right'
                label.ori = -45

        # Legacy
        if 'whiteOnBlack' in styleTweaks:
            self.line.color = 'black'
            self.tickLines.colors = 'black'
            self.marker.color = 'white'
            for label in self.labelObjs:
                label.color = 'white'


# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import
from __future__ import print_function

import os
import sys
from collections import deque
from past.builtins import unicode

try:
    import psychopy.logging as psycho_logging
    from psychopy.gui.qtgui import warnDlg
except ImportError:
    psycho_logging = None

from . import ioHubConnection
from ..util import yload, ydump, yLoader, yDumper
from ..errors import printExceptionDetailsToStdErr
from ..devices import Computer


######################## ioHubExperimentRuntime ##########################
# pylint: disable=line-too-long
# pylint: disable=superfluous-parens
# pylint: disable=no-self-use
# pylint: disable=global-statement
# pylint: disable=pointless-string-statement
# pylint: disable=protected-access
# pylint: disable=broad-except
# pylint: disable=fixme
# pylink: disable=unused-variable
# pylink: disable=too-many-instance-attributes

class ioHubExperimentRuntime(object):
    # TODO: Deprecate. Standardize on using launchHubServer function.
    # Not going to correct this classes code format (pylint, pep8)
    # since it is on its way out..
    """The ioHubExperimentRuntime class brings together several aspects of the
    ioHub Event Monitoring Framework, making it simpler to define and manage
    experiments that use multiple ioHub Device types, particularly when using
    more complicated devices such as the Eye Tracker or Analog Input Device.

    Other benefits of using the ioHubExperimentRuntime class include:

    * Automatic creation of an ioHubConnection instance with configuration of devices based on the associated Device Configuration Files.
    * Access to the ioHubConnection instance, all created ioHub Devices, and the ioHub Computer Device interface via two class attributes of the ioHubExperimentRuntime.
    * Optional support for the presentation of an Experiment Information Dialog at the start of each experiment session, based on the experiment settings specified in one of the associated configuration files.
    * Optional support for the presentation of a Session Variable Input Dialog at the start of each experiment session, based on the session settings specified in one of the associated configuration files. This includes the ability to collect input for custom experimenter defined session level fields that is stored in the ioHub DataStore for later retrieval and association with the event data collected during the experiment session.
    * Runtime access to the experiment, session, and device level configuration settings being used by the experiment in the form of python dictionary objects.
    * Automatic closure of the ioHub Process and the PsychoPy Process at the end of the experiment session, even when an unhandled exception occurs within your experiment scripting.

    The ioHubExperimentRuntime class is used to define the main Python script that
    will be run during each session of the experiment being run. In addition
    to the Python file containing the ioHubExperimentRuntime class extension,
    two configuration files are created:

    #. experiment_config.yaml : This file contains configuration details about the experiment itself, the experiment sessions that will be run to collect data from each participant of the experiment, and allows for process affinities to be set for the Experiment Process, ioHub Process, as well as all other processing on the computer. For details on defining an experiment_config.yaml file for use with the ioHubExperimentRuntime class, please see the Configuration Files section of the documentation.
    #. iohub_config.yaml : This file contains configuration details about each device that is being used by the experiment, as well as the ioHub DataStore. For details on defining an iohub_config.yaml file for use with the ioHubExperimentRuntime class, please see the Configuration Files section of the documentation.

    By separating experiment and session meta data definitions, as well as device
    configuration details, from the experiment paradigm logic contained within the
    ioHubExperimentRuntime class extension created, the ioHub Event Framework
    makes it possible to modify or switch between different implementations of an
    ioHub Device Interface without having to modify the experiment program logic.
    This is currently most beneficial when using an Eye Tracker or Analog Input
    Device, as these Device Interfaces support more than one hardware implementation.

    Many of the example scripts provided with the ioHub distribution use the
    ioHubExperimentRuntime class and config.yaml configuration files. The second
    example used in the Quick Start section of the documentation also uses this
    approach. Please refer to these resources for examples of using the
    ioHubExperimentRuntime class when creating an ioHub enabled project.

    Finally, there is an example called *startingTemplate* in the top level ioHub
    examples folder that contains a Python file with the base ioHubExperimentRuntime
    class extension in it, along with the two necessary configuration files.
    This example project folder can be copied to a directory of your choosing and renamed.
    Simply add the experiment logic you need to the run() method in the run.py file
    of the project, modify the experiment_config.yaml file to reflect the details of
    your intended application or experiment paradigm, and modify the iohub_config.yaml
    ensuring the devices required by your program are defined as needed. Then run the
    project by launching the run.py script with a Python interpreter.

    """

    def __init__(self, configFilePath, configFile):

        #: The hub attribute is the ioHubConnection class instance
        #: created for the ioHubExperimentRuntime. When the custom script
        #: provided in ioHubExperimentRuntime.run() is called, .hub is already
        #: set to an active ioHubConnection instance.
        self.hub = None

        #: The devices attribute is a short cut to the ioHubConnection
        #: instance's .devices attribute. i.e. self.devices = self.hub.devices.
        #: A reference to the Computer class is also added to the devices
        #: attribute, so when using the ioHubConnection devices attribute,
        #: the ioHub Computer class can be accessed using self.devices.computer;
        #: It does not need to be imported by your script.
        self.devices = None

        self.configFilePath = configFilePath
        self.configFileName = configFile

        # load the experiment config settings from the experiment_config.yaml file.
        # The file must be in the same directory as the experiment script.
        self.configuration = yload(
            open(
                os.path.join(
                    self.configFilePath,
                    self.configFileName),
                u'r'),
            Loader=yLoader)

        import random
        random.seed(Computer.getTime() * 1000.123)
        randomInt = random.randint(1, 1000)
        self.experimentConfig = dict()
        self._experimentConfigKeys = [
            'title', 'code', 'version', 'description']
        self.experimentConfig.setdefault(
            'title', self.experimentConfig.get(
                'title', 'A Default Experiment Title'))
        self.experimentConfig.setdefault(
            'code', self.experimentConfig.get(
                'code', 'Def_Exp_Code'))
        self.experimentConfig.setdefault(
            'version', self.experimentConfig.get(
                'version', '0.0.0'))
        self.experimentConfig.setdefault(
            'description', self.experimentConfig.get(
                'description', 'A Default Experiment Description'))
#        self.experimentConfig.setdefault('total_sessions_to_run',self.experimentConfig.get('total_sessions_to_run',0))

        for key in self._experimentConfigKeys:
            if key in self.configuration:
                self.experimentConfig[key] = self.configuration[key]

        self.experimentSessionDefaults = self.configuration.get(
            'session_defaults', {})
        self.sessionUserVariables = self.experimentSessionDefaults.get(
            'user_variables', None)
        if self.sessionUserVariables is not None:
            del self.experimentSessionDefaults['user_variables']
        else:
            self.sessionUserVariables = {}

        # initialize the experiment object based on the configuration settings.
        self.hub = self._initalizeConfiguration()

        self.devices = self.hub.devices
        self.devices.computer = Computer

    def run(self, *sys_argv):
        """The run method must be overwritten by your subclass of
        ioHubExperimentRuntime, and would include the equivalent logic to what
        would be added to the main starting script in a procedural PsychoPy
        script.

        When the run method starts, the ioHub Server is online and any devices
        specified for the experiment are ready for use. When the contents of the run method
        allow the method to return or end, the experiment session is complete.

        Any sys_argv are equal to the sys.argv received by the script when it was started.

        Args:
            sys_argv (list): The list of arguments passed to the script when it was started with Python.

        Returns:
            User defined.

        """
        pass

    def getConfiguration(self):
        """Returns the full parsing of experiment_config.yaml as a python
        dictionary.

        Args:
            None

        Returns:
            dict: The python object representation of the contents of the experiment_config.yaml file loaded for the experiment.

        """
        return self.configuration

    def getExperimentMetaData(self):
        """
        Returns the experiment parameters saved to the ioHub DataStore experiment_metadata table.
        The values are actually only saved the first time the experiment is run.
        The variable names and values contained within the returned dict are also what
        would be presented at the experiment start in the read-only Experiment Information Dialog.

        Args:
            None

        Returns:
            dict: The python object representation of the experiment meta data, namely the experiment_code, title, version, and description fields.
        """
        if self.hub is not None:
            return self.hub.getExperimentMetaData()
        return self.experimentConfig

    def getSessionMetaData(self):
        """Returns the experiment session parameters saved to the ioHub
        DataStore for the current experiment session. These are the parameters
        defined in the session_defaults section of the experiment_config.yaml
        and are also optionally displayed in the Session Input Dialog at the
        start of each experiment session.

        Args:
            None

        Returns:
            dict: The python object representation of the session meta data saved to the ioHub DataStore for the current experiment run.

        """
        if self.hub is not None:
            return self.hub.getSessionMetaData()
        return self.experimentSessionDefaults

    def getUserDefinedParameters(self):
        """Return only the user defined session parameters defined in the
        experiment_config.yaml. These parameters are displayed in the Session
        Input Dialog (if enabled) and the value entered for each parameter is
        provide in the state of the returned dict. These parameters and values
        are also saved in the session meta data table of the ioHub DataStore.

        Args:
            None

        Returns:
            dict: The python object representation of the user defined session parameters saved to the ioHub DataStore for the current experiment run.

        """
        return self.sessionUserVariables

    def isSessionCodeInUse(self, current_sess_code):
        """Session codes must be unique within an experiment. This method will
        return True if the provided session code is already used in one of the
        existing experiment sessions saved to the ioHub DataStore. False is
        returned if the session code is not used, and would therefore make a
        valid session code for the current run.

        Args:
            current_sess_code (str): The string being requested to be used as the
            current experiment session code. maximum length is 24 characters.

        Returns:
            bool: True if the code given is already in use. False if it is not in use.

        """
        r = self.hub._sendToHubServer(
            ('RPC', 'checkIfSessionCodeExists', (current_sess_code,)))
        return r[2]

    def prePostExperimentVariableCallback(self, experiment_meta_data):
        """This method is called prior to the experiment meta data being sent
        to the ioHub DataStore to be saved as the details regarding the current
        experiment being run. Any changes made to the experiment_meta_data dict
        passed into the method will be reflected in the data values saved to
        the ioHub DataStore.

        Note that the same dict object that is passed into the method as an argument
        must be returned by the method as the result.

        Args:
            experiment_meta_data (dict): The state of the experiment meta data prior to being sent to the ioHub DataStore for storage.

        Returns:
            dict: The experiment_meta_data arg passed to the method.

        """
        return experiment_meta_data

    def prePostSessionVariableCallback(self, session_meta_data):
        """This method is called prior to the session meta data being sent to
        the ioHub DataStore to be saved as the details regarding the current
        session being run. Any changes made to the session_meta_data dict
        passed into the method will be reflected in the data values saved to
        the ioHub DataStore for the session.

        Note that the same dict object that is passed into the method as an argument
        must be returned by the method as the result.

        Args:
            session_meta_data (dict): The state of the session meta data prior to being sent to the ioHub DataStore for storage.

        Returns:
            dict: The session_meta_data arg passed to the method.

        """
        org_sess_code = session_meta_data.setdefault('code', 'default_sess')
        scount = 1
        sess_code = org_sess_code
        while self.isSessionCodeInUse(sess_code) is True:
            sess_code = '%s-%d' % (org_sess_code, scount)
            scount += 1
        session_meta_data['code'] = sess_code
        return session_meta_data

    @staticmethod
    def printExceptionDetails():
        """Prints out stack trace information for the last exception raised by
        the PsychoPy Process.

        Currently a lot of redundant data is printed regarding the exception and stack trace.

        TO DO: clean this up so there is not so much redundant info printed.

        Args:
            None

        Returns:
            None

        """
        import traceback
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print('*** print_tb:')
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print('*** print_exception:')
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                                  limit=2, file=sys.stdout)
        print('*** print_exc:')
        traceback.print_exc()
        print('*** format_exc, first and last line:')
        formatted_lines = traceback.format_exc().splitlines()
        print(formatted_lines[0])
        print(formatted_lines[-1])
        print('*** format_exception:')
        print(repr(traceback.format_exception(exc_type, exc_value,
                                              exc_traceback)))
        print('*** extract_tb:')
        print(repr(traceback.extract_tb(exc_traceback)))
        print('*** format_tb:')
        print(repr(traceback.format_tb(exc_traceback)))
        print(('*** tb_lineno:', exc_traceback.tb_lineno))

    @staticmethod
    def mergeConfigurationFiles(
            base_config_file_path,
            update_from_config_file_path,
            merged_save_to_path):
        """Merges two iohub configuration files into one and saves it to a file
        using the path/file name in merged_save_to_path."""
        base_config = yload(open(base_config_file_path, 'r'), Loader=yLoader)
        update_from_config = yload(
            open(
                update_from_config_file_path,
                'r'),
            Loader=yLoader)

        def merge(update, base):
            if isinstance(update, dict) and isinstance(base, dict):
                for k, v in base.items():
                    if k not in update:
                        update[k] = v
                    else:
                        if isinstance(update[k], list):
                            if isinstance(v, list):
                                v.extend(update[k])
                                update[k] = v
                            else:
                                update[k].insert(0, v)
                        else:
                            update[k] = merge(update[k], v)
            return update

        import copy
        merged = merge(copy.deepcopy(update_from_config), base_config)
        ydump(merged, open(merged_save_to_path, 'w'), Dumper=yDumper)

        return merged

    def _initalizeConfiguration(self):
        global _currentSessionInfo
        """
        Based on the configuration data in the experiment_config.yaml and iohub_config.yaml,
        configure the experiment environment and ioHub process environments. This mehtod is called by the class init
        and should not be called directly.
        """
        display_experiment_dialog = self.configuration.get(
            'display_experiment_dialog', False)
        display_session_dialog = self.configuration.get(
            'display_session_dialog', False)

        if display_experiment_dialog is True:
            # display a read only dialog verifying the experiment parameters
            # (based on the experiment .yaml file) to be run. User can hit OK to continue,
            # or Cancel to end the experiment session if the wrong experiment
            # was started.
            exitExperiment = self._displayExperimentSettingsDialog()
            if exitExperiment:
                print('User Cancelled Experiment Launch.')
                self._close()
                sys.exit(1)

        self.experimentConfig = self.prePostExperimentVariableCallback(
            self.experimentConfig)

        ioHubInfo = self.configuration.get('ioHub', {})

        if ioHubInfo is None:
            print('ioHub section of configuration file could not be found. Exiting.....')
            self._close()
            sys.exit(1)
        else:
            ioHubConfigFileName = unicode(
                ioHubInfo.get('config', 'iohub_config.yaml'))
            ioHubConfigAbsPath = os.path.join(
                self.configFilePath, unicode(ioHubConfigFileName))
            self.hub = ioHubConnection(None, ioHubConfigAbsPath)

            # print 'ioHubExperimentRuntime.hub: {0}'.format(self.hub)
            # A circular buffer used to hold events retrieved from self.getEvents() during
            # self.delay() calls. self.getEvents() appends any events in the allEvents
            # buffer to the result of the hub.getEvents() call that is made.
            self.hub.allEvents = deque(
                maxlen=self.configuration.get(
                    'event_buffer_length', 256))

            # print 'ioHubExperimentRuntime sending experiment config.....'
            # send experiment info and set exp. id
            self.hub._sendExperimentInfo(self.experimentConfig)

            # print 'ioHubExperimentRuntime SENT experiment config.'

            allSessionDialogVariables = dict(
                self.experimentSessionDefaults,
                **self.sessionUserVariables)
            sessionVariableOrder = self.configuration.get(
                'session_variable_order', [])
            if 'user_variables' in allSessionDialogVariables:
                del allSessionDialogVariables['user_variables']

            if display_session_dialog is True:
                # display session dialog
                r = True
                while r is True:
                    # display editable session variable dialog displaying the ioHub required session variables
                    # and any user defined session variables (as specified in the experiment config .yaml file)
                    # User can enter correct values and hit OK to continue, or
                    # Cancel to end the experiment session.

                    allSessionDialogVariables = dict(
                        self.experimentSessionDefaults, **self.sessionUserVariables)
                    sessionVariableOrder = self.configuration.get(
                        'session_variable_order', [])
                    if 'user_variables' in allSessionDialogVariables:
                        del allSessionDialogVariables['user_variables']

                    tempdict = self._displayExperimentSessionSettingsDialog(
                        allSessionDialogVariables, sessionVariableOrder)
                    if tempdict is None:
                        print('User Cancelled Experiment Launch.')
                        self._close()
                        sys.exit(1)

                    tempdict['user_variables'] = self.sessionUserVariables

                    r = self.isSessionCodeInUse(tempdict['code'])

                    if r is True:
                        display_device = self.hub.getDevice('display')
                        display_id = 0
                        if display_device:
                            display_id = display_device.getIndex()
                        msg_dialog = warnDlg('Session Code In Use',
                                             'Session Code {0} is already in '
                                             'use by the experiment.\nPlease '
                                             'enter a new Session '
                                             'Code'.format(tempdict['code']))
                        msg_dialog.show()
            else:
                tempdict = allSessionDialogVariables
                tempdict['user_variables'] = self.sessionUserVariables

            for key, value in allSessionDialogVariables.items():
                if key in self.experimentSessionDefaults:
                    # (u''+value).encode('utf-8')
                    self.experimentSessionDefaults[key] = value
                elif key in self.sessionUserVariables:
                    # (u''+value).encode('utf-8')
                    self.sessionUserVariables[key] = value

            tempdict = self.prePostSessionVariableCallback(tempdict)
            tempdict['user_variables'] = self.sessionUserVariables

            _currentSessionInfo = self.experimentSessionDefaults

            self.hub._sendSessionInfo(tempdict)

            self._setInitialProcessAffinities(ioHubInfo)

            return self.hub

    def _setInitialProcessAffinities(self, ioHubInfo):
            # set process affinities based on config file settings
        cpus = range(Computer.processing_unit_count)
        experiment_process_affinity = cpus
        other_process_affinity = cpus
        iohub_process_affinity = cpus

        experiment_process_affinity = self.configuration.get(
            'process_affinity', [])
        if len(experiment_process_affinity) == 0:
            experiment_process_affinity = cpus

        other_process_affinity = self.configuration.get(
            'remaining_processes_affinity', [])
        if len(other_process_affinity) == 0:
            other_process_affinity = cpus

        iohub_process_affinity = ioHubInfo.get('process_affinity', [])
        if len(iohub_process_affinity) == 0:
            iohub_process_affinity = cpus

        if len(experiment_process_affinity) < len(
                cpus) and len(iohub_process_affinity) < len(cpus):
            Computer.setProcessAffinities(
                experiment_process_affinity,
                iohub_process_affinity)

        if len(other_process_affinity) < len(cpus):
            ignore = [Computer.currentProcessID, Computer.iohub_process_id]
            Computer.setAllOtherProcessesAffinity(
                other_process_affinity, ignore)

    def start(self, *sys_argv):
        """This method should be called from within a user script which as
        extended this class to start the ioHub Server. The run() method of the
        class, containing the user experiment logic, is then called. When the
        run() method completes, the ioHub Server is stopped and the program
        exits.

        Args: None
        Return: None

        """
        try:
            result = self.run(*sys_argv)
            self._close()
            return result
        except Exception:
            printExceptionDetailsToStdErr()
            self._close()

    def _displayExperimentSettingsDialog(self):
        """
        Display a read-only dialog showing the experiment setting retrieved from the configuration file. This gives the
        experiment operator a chance to ensure the correct configuration file was loaded for the script being run. If OK
        is selected in the dialog, the experiment logic continues, otherwise the experiment session is terminated.
        """
        # print 'self.experimentConfig:', self.experimentConfig
        # print 'self._experimentConfigKeys:',self._experimentConfigKeys
        result = True
        try:
            from psychopy import gui
            experimentDlg = gui.DlgFromDict(
                self.experimentConfig,
                'Experiment Launcher',
                self._experimentConfigKeys,
                self._experimentConfigKeys,
                {})
            if experimentDlg.OK:
                result = False
            else:
                result = True
        except ImportError:
            result = False
        return result

    def _displayExperimentSessionSettingsDialog(
            self, allSessionDialogVariables, sessionVariableOrder):
        """Display an editable dialog showing the experiment session setting
        retrieved from the configuration file.

        This includes the few mandatory ioHub experiment session
        attributes, as well as any user defined experiment session
        attributes that have been defined in the experiment
        configuration file. If OK is selected in the dialog, the
        experiment logic continues, otherwise the experiment session is
        terminated.

        """
        result = None
        try:
            from psychopy import gui
            sessionDlg = gui.DlgFromDict(
                allSessionDialogVariables,
                'Experiment Session Settings',
                [],
                sessionVariableOrder)
            if sessionDlg.OK:
                result = allSessionDialogVariables
        except ImportError:
            result = None
        return result

    def _close(self):
        """Close the experiment runtime and the ioHub server process."""
        # terminate the ioServer
        if self.hub:
            self.hub._shutDownServer()
        # terminate psychopy
        # core.quit()

    def __del__(self):
        try:
            if self.hub:
                self.hub._shutDownServer()
        except Exception:
            pass
        self.hub = None
        self.devices = None

# pylint: enable=line-too-long
# pylint: enable=superfluous-parens
# pylint: enable=no-self-use
# pylint: enable=global-statement
# pylint: enable=pointless-string-statement
# pylint: enable=protected-access
# pylint: enable=broad-except
# pylint: enable=fixme
# pylink: enable=unused-variable
# pylink: enable=too-many-instance-attributes



# -*- coding: utf-8 -*-
"""Classes for color picker dialog panels."""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import wx
from wx.lib.buttons import GenButton
from wx.lib.scrolledpanel import ScrolledPanel
from psychopy.colors import Color, colorNames
import numpy as np


class ColorPresets(ScrolledPanel):
    """Class for creating a scrollable button list that displays all preset
    colors.

    Parameters
    ----------
    parent : object
        Object this panel belongs to (i.e. :class:`wx.Frame` or `wx.Panel`).

    """
    def __init__(self, parent):
        # originally made by Todd Parsons
        ScrolledPanel.__init__(
            self,
            parent,
            size=wx.Size(120, -1),
            style=wx.VSCROLL | wx.BORDER_NONE)

        self.parent = parent
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self._createColorButtons()
        self.SetSizer(self.sizer)
        self.SetupScrolling()

    def _createColorButtons(self):
        """Generate color buttons based on the presets defined in the `colors`
        module.

        When a user clicks on the buttons, it changes the current color the
        colorspace page is displaying.

        """
        # create buttons for each preset color
        colorList = list(colorNames)
        btnSize = wx.Size(120, 24)
        for color in colorList:
            btn = GenButton(self, size=btnSize, label=color, name=color)
            btn.colorData = col = Color(color, 'named')
            btn.SetOwnBackgroundColour(col.rgba255)

            # Compute the (perceived) luminance of the color, used to set the
            # foreground text to ensure it's legible on the background. Uses the
            # the luminance part of formula to convert RGB1 to YIQ.
            luminance = np.sum(np.asarray((0.299, 0.587, 0.114)) * col.rgb1)
            if luminance < 0.5:
                btn.SetForegroundColour(wx.WHITE)
            else:
                btn.SetForegroundColour(wx.BLACK)

            btn.SetBezelWidth(0)
            btn.SetUseFocusIndicator(False)
            btn.Bind(wx.EVT_BUTTON, self.onClick)
            self.sizer.Add(btn, 1, wx.ALL | wx.EXPAND, 0)

    def onClick(self, event):
        """Event called when the user clicks a color button. Value is passed to
        the dialog and updates the color. This in turn will update the values of
        all the color space pages.

        """
        self.GetTopLevelParent().color = event.GetEventObject().colorData.copy()
        event.Skip()


class ColorPreview(wx.Panel):
    """Class for the color preview panel in the color picker.

    This panel displays the current color specified by the user. A background
    checkerboard pattern is drawn as a background making transparency more
    apparent.

    """
    def __init__(self, parent, color):
        wx.Panel.__init__(self, parent, size=(100, -1))
        # device contexts for drawing
        self.pdc = self.dc = None

        self.parent = parent
        self.SetDoubleBuffered(True)
        self.color = color
        self.Bind(wx.EVT_PAINT, self.onPaint)

    @property
    def color(self):
        """Color being displayed (:class:`~psychopy.colors.Color`)."""
        return self._color

    @color.setter
    def color(self, value):
        self._color = value
        self.Refresh()

    def onPaint(self, event):
        """Called each time the preview is updated or `color` is changed. Bound
        to `EVT_PAINT`. The background is only drawn if the color is
        transparent.

        """
        self.pdc = wx.PaintDC(self)
        self.dc = wx.GCDC(self.pdc)

        # only draw background if there is transparency, reduces draw calls
        if self._color.alpha < 1.0:
            self._paintCheckerboard()

        self._paintPreviewColor()

    def _paintPreviewColor(self):
        """Paint the current color. Called when `onPaint` is invoked, but after
        the checkerboard is drawn.

        """
        if self.dc is None:
            return  # nop

        # originally written by Todd Parsons
        self.dc.SetBrush(
            wx.Brush(list(self.color.rgb255) + [self.color.alpha * 255],
                     wx.BRUSHSTYLE_TRANSPARENT))
        self.dc.SetPen(
            wx.Pen(list(self.color.rgb255) + [self.color.alpha * 255],
                   wx.PENSTYLE_TRANSPARENT))
        self.dc.DrawRectangle(0, 0, self.GetSize()[0], self.GetSize()[1])

    def _paintCheckerboard(self, gridRes=10):
        """Paint the background checkerboard grid of the color preview area.
        this provides a background to make the effect of adjusting the alpha
        channel more apparent.

        Must be called within the `onPaint` method only when the color is
        transparent. Don't call elsewhere.

        Parameters
        ----------
        gridRes : int
            Width and height of each grid square.

        """
        if self.pdc is None:
            return  # nop

        self.pdc.SetBrush(wx.LIGHT_GREY_BRUSH)
        self.pdc.SetPen(wx.LIGHT_GREY_PEN)

        # originally written by Todd Parsons
        w = h = gridRes
        for x in range(0, self.GetSize()[0], w * 2):
            for y in range(0 + (x % 2) * h, self.GetSize()[1], h * 2):
                self.pdc.DrawRectangle(x, y, w, h)
                self.pdc.DrawRectangle(x + w, y + h, w, h)


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Utility functions to support Experiment classes
"""
import re

# this needs to be accessed from __str__ method of Param
scriptTarget = "PsychoPy"

# predefine some regex's; deepcopy complains if do in NameSpace.__init__()
unescapedDollarSign_re = re.compile(r"^\$|[^\\]\$")  # detect "code wanted"
valid_var_re = re.compile(r"^[a-zA-Z_][\w]*$")  # filter for legal var names
nonalphanumeric_re = re.compile(r'\W')  # will match all bad var name chars


class CodeGenerationException(Exception):
    """
    Exception thrown by a component when it is unable to generate its code.
    """

    def __init__(self, source, message=""):
        super(CodeGenerationException, self).__init__()
        self.source = source
        self.message = message

    def __str__(self):
        return "{}: ".format(self.source, self.message)



#!/usr/bin/env python

# Copyright (c) 1996-2002 Denis G. Pelli
# Copyright (c) 1996-9 David Brainard
# Copyright (c) 2004-7 Andrew D. Straw
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the Enthought nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from __future__ import absolute_import, division, print_function
from past.builtins import raw_input

from builtins import zip
from builtins import range
from builtins import object
__all__ = ['QuestObject']

import math
import copy
import warnings
import random
import sys
import time

import numpy as num

def getinf(x):
    return num.nonzero( num.isinf( num.atleast_1d(x) ) )


class QuestObject(object):

    """Measure threshold using a Weibull psychometric function.

    Threshold 't' is measured on an abstract 'intensity' scale, which
    usually corresponds to log10 contrast.

    The Weibull psychometric function:

    p2=delta*gamma+(1-delta)*(1-(1-gamma)*exp(-10**(beta*(x2+xThreshold))))

    where x2 represents log10 intensity relative to threshold
    (i.e., x2 = x - T, where x is intensity, and T is threshold intensity).
    xThreshold shifts the psychometric function along the intensity axis
    such that threshold performance (specified as pThreshold below) will
    occur at intensity x = T, i.e., x2 = x - T = 0. In the 
    Watson & Pelli (1983) paper, xThreshold is denoted as epsilon and used
    to perform testing at the "ideal sweat factor".
    The Weibull function itself appears only in recompute(), which uses
    the specified parameter values in self to compute a psychometric
    function and store it in self. All the other methods simply use
    the psychometric function stored as instance
    variables. recompute() is called solely by __init__() and
    beta_analysis() (and possibly by a few user programs). Thus, if
    you prefer to use a different kind of psychometric function,
    called Foo, you need only subclass QuestObject, overriding
    __init__(), recompute(), and (if you need it) beta_analysis().

    instance variables:

    tGuess is your prior threshold estimate.

    tGuessSd is the standard deviation you assign to that guess.

    pThreshold is your threshold criterion expressed as probability of
    response==1. An intensity offset is introduced into the
    psychometric function so that threshold (i.e. the midpoint of the
    table) yields pThreshold.

    beta, delta, and gamma are the parameters of a Weibull
    psychometric function.

    beta controls the steepness of the psychometric
    function. Typically 3.5.

    delta is the fraction of trials on which the observer presses
    blindly.  Typically 0.01.

    gamma is the fraction of trials that will generate response 1 when
    intensity==-inf.

    grain is the quantization of the internal table. E.g. 0.01.

    range is the intensity difference between the largest and smallest
    intensity that the internal table can store. E.g. 5. This interval
    will be centered on the initial guess tGuess,
    i.e. [tGuess-range/2, tGuess+range/2].  QUEST assumes that
    intensities outside of this interval have zero prior probability,
    i.e. they are impossible.

    """
    def __init__(self,tGuess,tGuessSd,pThreshold,beta,delta,gamma,grain=0.01,range=None):
        """Initialize Quest parameters.

        Create an instance of QuestObject with all the information
        necessary to measure threshold.

        This was converted from the Psychtoolbox's QuestCreate function.
        """
        super(QuestObject, self).__init__()
        grain = float(grain) # make sure grain is a float
        if range is None:
            dim = 500
        else:
            if range <= 0:
                raise ValueError('argument "range" must be greater than zero.')
            dim=range/grain
            dim=2*math.ceil(dim/2.0) # round up to even integer
        self.updatePdf = True
        self.warnPdf = True
        self.normalizePdf = False
        self.tGuess = tGuess
        self.tGuessSd = tGuessSd
        self.pThreshold = pThreshold
        self.beta = beta
        self.delta = delta
        self.gamma = gamma
        self.grain = grain
        self.dim = dim
        self.recompute()

    def beta_analysis(self,stream=None):
        """Analyze the quest function with beta as a free parameter.

        It returns the mean estimates of alpha (as logC) and
        beta. Gamma is left at whatever value the user fixed it at.
        """

        def _beta_analysis1(stream=None):
            """private function called by beta_analysis()"""
            if stream is None:
                stream=sys.stdout
            q2 = []
            for i in range(1,17):
                q_copy=copy.copy(self)
                q_copy.beta=2**(i/4.0)
                q_copy.dim=250
                q_copy.grain=0.02
                q_copy.recompute()
                q2.append(q_copy)
            na = num.array # shorthand
            t2    = na([q2i.mean() for q2i in q2])
            p2    = na([q2i.pdf_at(t2i) for q2i,t2i in zip(q2,t2)])
            sd2   = na([q2i.sd() for q2i in q2])
            beta2 = na([q2i.beta for q2i in q2])
            i=num.argsort(p2)[-1]
            t=t2[i]
            sd=q2[i].sd()
            p=num.sum(p2)
            betaMean=num.sum(p2*beta2)/p
            betaSd=math.sqrt(num.sum(p2*beta2**2)/p-(num.sum(p2*beta2)/p)**2)
            iBetaMean=num.sum(p2/beta2)/p
            iBetaSd=math.sqrt(num.sum(p2/beta2**2)/p-(num.sum(p2/beta2)/p)**2)
            stream.write('%5.2f	%5.2f	%4.1f	%4.1f	%6.3f\n'%(t, sd, 1/iBetaMean, betaSd, self.gamma))
        print('Now re-analyzing with beta as a free parameter. . . .')
        if stream is None:
            stream=sys.stdout
        stream.write('logC 	 sd 	 beta	 sd	 gamma\n');
        _beta_analysis1(stream)

    def mean(self):
        """Mean of Quest posterior pdf.

        Get the mean threshold estimate.

        This was converted from the Psychtoolbox's QuestMean function.
        """
        return self.tGuess + num.sum(self.pdf*self.x)/num.sum(self.pdf)

    def mode(self):
        """Mode of Quest posterior pdf.

        t,p=q.mode()
        't' is the mode threshold estimate
        'p' is the value of the (unnormalized) pdf at t.

        This was converted from the Psychtoolbox's QuestMode function.
        """
        iMode = num.argsort(self.pdf)[-1]
        p=self.pdf[iMode]
        t=self.x[iMode]+self.tGuess
        return t,p

    def p(self,x):
        """probability of correct response at intensity x.

        p=q.p(x)

        The probability of a correct (or yes) response at intensity x,
        assuming threshold is at x=0.

        This was converted from the Psychtoolbox's QuestP function.
        """
        if x < self.x2[0]:
            return self.x2[0]
        if x > self.x2[-1]:
            return self.x2[-1]
        return num.interp(x,self.x2,self.p2)

    def pdf_at(self,t):
        """The (unnormalized) probability density of candidate threshold 't'.

        This was converted from the Psychtoolbox's QuestPdf function.
        """
        i=int(round((t-self.tGuess)/self.grain))+1+self.dim/2
        i=min(len(self.pdf),max(1,i))-1
        p=self.pdf[i]
        return p

    def quantile(self,quantileOrder=None):
        """Get Quest recommendation for next trial level.

        intensity=q.quantile([quantileOrder])

        Gets a quantile of the pdf in the struct q.  You may specify
        the desired quantileOrder, e.g. 0.5 for median, or, making two
        calls, 0.05 and 0.95 for a 90confidence interval.  If the
        'quantileOrder' argument is not supplied, then it's taken from
        the QuestObject instance. __init__() uses recompute() to
        compute the optimal quantileOrder and saves that in the
        QuestObject instance; this quantileOrder yields a quantile
        that is the most informative intensity for the next trial.

        This was converted from the Psychtoolbox's QuestQuantile function.
        """
        if quantileOrder is None:
            quantileOrder = self.quantileOrder
        p = num.cumsum(self.pdf)
        if len(getinf(p[-1])[0]):
            raise RuntimeError('pdf is not finite')
        if p[-1]==0:
            raise RuntimeError('pdf is all zero')
        m1p = num.concatenate(([-1],p))
        index = num.nonzero( m1p[1:]-m1p[:-1] )[0]
        if len(index) < 2:
            raise RuntimeError('pdf has only %g nonzero point(s)'%len(index))
        ires = num.interp([quantileOrder*p[-1]],p[index],self.x[index])[0]
        return self.tGuess+ires

    def sd(self):
        """Standard deviation of Quest posterior pdf.

        Get the sd of the threshold distribution.

        This was converted from the Psychtoolbox's QuestSd function."""
        p=num.sum(self.pdf)
        sd=math.sqrt(num.sum(self.pdf*self.x**2)/p-(num.sum(self.pdf*self.x)/p)**2)
        return sd

    def simulate(self,tTest,tActual):
        """Simulate an observer with given Quest parameters.

        response=QuestSimulate(q,intensity,tActual)

        Simulate the response of an observer with threshold tActual.

        This was converted from the Psychtoolbox's QuestSimulate function."""
        t = min( max(tTest-tActual, self.x2[0]), self.x2[-1] )
        response= num.interp([t],self.x2,self.p2)[0] > random.random()
        return response

    def recompute(self):
        """Recompute the psychometric function & pdf.

        Call this immediately after changing a parameter of the
        psychometric function. recompute() uses the specified
        parameters in 'self' to recompute the psychometric
        function. It then uses the newly computed psychometric
        function and the history in self.intensity and self.response
        to recompute the pdf. (recompute() does nothing if q.updatePdf
        is False.)

        This was converted from the Psychtoolbox's QuestRecompute function."""
        if not self.updatePdf:
            return
        if self.gamma > self.pThreshold:
            warnings.warn( 'reducing gamma from %.2f to 0.5'%self.gamma)
            self.gamma = 0.5
        self.i = num.arange(-self.dim/2, self.dim/2+1)
        self.x = self.i * self.grain
        self.pdf = num.exp(-0.5*(self.x/self.tGuessSd)**2)
        self.pdf = self.pdf/num.sum(self.pdf)
        i2 = num.arange(-self.dim,self.dim+1)
        self.x2 = i2*self.grain
        self.p2 = self.delta*self.gamma+(1-self.delta)*(1-(1-self.gamma)*num.exp(-10**(self.beta*self.x2)))
        if self.p2[0] >= self.pThreshold or self.p2[-1] <= self.pThreshold:
            raise RuntimeError('psychometric function range [%.2f %.2f] omits %.2f threshold'%(self.p2[0],self.p2[-1],self.pThreshold)) # XXX
        if len(getinf(self.p2)[0]):
            raise RuntimeError('psychometric function p2 is not finite')
        index = num.nonzero( self.p2[1:]-self.p2[:-1] )[0] # strictly monotonic subset
        if len(index) < 2:
            raise RuntimeError('psychometric function has only %g strictly monotonic points'%len(index))
        self.xThreshold = num.interp([self.pThreshold],self.p2[index],self.x2[index])[0]
        self.p2 = self.delta*self.gamma+(1-self.delta)*(1-(1-self.gamma)*num.exp(-10**(self.beta*(self.x2+self.xThreshold))))
        if len(getinf(self.p2)[0]):
            raise RuntimeError('psychometric function p2 is not finite')
        self.s2 = num.array( ((1-self.p2)[::-1], self.p2[::-1]) )
        if not hasattr(self,'intensity') or not hasattr(self,'response'):
            self.intensity = []
            self.response = []
        if len(getinf(self.s2)[0]):
            raise RuntimeError('psychometric function s2 is not finite')

        eps = 1e-14

        pL = self.p2[0]
        pH = self.p2[-1]
        pE = pH*math.log(pH+eps)-pL*math.log(pL+eps)+(1-pH+eps)*math.log(1-pH+eps)-(1-pL+eps)*math.log(1-pL+eps)
        pE = 1/(1+math.exp(pE/(pL-pH)))
        self.quantileOrder=(pE-pL)/(pH-pL)

        if len(getinf(self.pdf)[0]):
            raise RuntimeError('prior pdf is not finite')

        # recompute the pdf from the historical record of trials
        for intensity, response in zip(self.intensity, self.response):
            inten = max(-1e10,min(1e10, intensity)) # make intensity finite
            ii = len(self.pdf) + self.i-round((inten-self.tGuess)/self.grain)-1
            if ii[0]<0:
                ii = ii-ii[0]
            if ii[-1]>=self.s2.shape[1]:
                ii = ii+self.s2.shape[1]-ii[-1]-1
            iii = ii.astype(num.int_)
            if not num.allclose(ii,iii):
                raise ValueError('truncation error')
            self.pdf = self.pdf*self.s2[response,iii]
            if self.normalizePdf and ii % 100 == 0:
                self.pdf = self.pdf/num.sum(self.pdf) # avoid underflow; keep the pdf normalized
        if self.normalizePdf:
            self.pdf = self.pdf/num.sum(self.pdf) # avoid underflow; keep the pdf normalized
        if len(getinf(self.pdf)[0]):
            raise RuntimeError('prior pdf is not finite')

    def update(self,intensity,response):
        """Update Quest posterior pdf.

        Update self to reflect the results of this trial. The
        historical records self.intensity and self.response are always
        updated, but self.pdf is only updated if self.updatePdf is
        true. You can always call QuestRecompute to recreate q.pdf
        from scratch from the historical record.

        This was converted from the Psychtoolbox's QuestUpdate function."""

        if response < 0 or response > self.s2.shape[0]:
            raise RuntimeError('response %g out of range 0 to %d'%(response,self.s2.shape[0]))
        if self.updatePdf:
            inten = max(-1e10,min(1e10,intensity)) # make intensity finite
            ii = len(self.pdf) + self.i-round((inten-self.tGuess)/self.grain)-1
            if ii[0]<0 or ii[-1] > self.s2.shape[1]:
                if self.warnPdf:
                    low=(1-len(self.pdf)-self.i[0])*self.grain+self.tGuess
                    high=(self.s2.shape[1]-len(self.pdf)-self.i[-1])*self.grain+self.tGuess
                    warnings.warn( 'intensity %.2f out of range %.2f to %.2f. Pdf will be inexact.'%(intensity,low,high),
                                   RuntimeWarning,stacklevel=2)
                if ii[0]<0:
                    ii = ii-ii[0]
                else:
                    ii = ii+self.s2.shape[1]-ii[-1]-1
            iii = ii.astype(num.int_)
            if not num.allclose(ii,iii):
                raise ValueError('truncation error')
            self.pdf = self.pdf*self.s2[response,iii]
            if self.normalizePdf:
                self.pdf=self.pdf/num.sum(self.pdf)
        # keep a historical record of the trials
        self.intensity.append(intensity)
        self.response.append(response)

def demo():
    """Demo script for Quest routines.

    By commenting and uncommenting a few lines in this function, you
    can use this file to implement three QUEST-related procedures for
    measuring threshold.

    QuestMode: In the original algorithm of Watson & Pelli (1983) each
    trial and the final estimate are at the MODE of the posterior pdf.

    QuestMean: In the improved algorithm of King-Smith et al. (1994).
    each trial and the final estimate are at the MEAN of the posterior
    pdf.

    QuestQuantile & QuestMean: In the ideal algorithm of Pelli (1987)
    each trial is at the best QUANTILE, and the final estimate is at
    the MEAN of the posterior pdf.

    This was converted from the Psychtoolbox's QuestDemo function.

    King-Smith, P. E., Grigsby, S. S., Vingrys, A. J., Benes, S. C.,
    and Supowit, A.  (1994) Efficient and unbiased modifications of
    the QUEST threshold method: theory, simulations, experimental
    evaluation and practical implementation.  Vision Res, 34 (7),
    885-912.

    Pelli, D. G. (1987) The ideal psychometric
    procedure. Investigative Ophthalmology & Visual Science, 28
    (Suppl), 366.

    Watson, A. B. and Pelli, D. G. (1983) QUEST: a Bayesian adaptive
    psychometric method. Percept Psychophys, 33 (2), 113-20.
    """

    print('The intensity scale is abstract, but usually we think of it as representing log contrast.')

    tActual = None
    while tActual is None:
        sys.stdout.write('Specify true threshold of simulated observer: ')
        input = raw_input()
        try:
            tActual = float(input)
        except Exception:
            pass

    tGuess = None
    while tGuess is None:
        sys.stdout.write('Estimate threshold: ')
        input = raw_input()
        try:
            tGuess = float(input)
        except Exception:
            pass

    tGuessSd = 2.0 # sd of Gaussian before clipping to specified range
    pThreshold = 0.82
    beta = 3.5
    delta = 0.01
    gamma = 0.5
    q=QuestObject(tGuess,tGuessSd,pThreshold,beta,delta,gamma)

    # Simulate a series of trials.
    trialsDesired=100
    wrongRight = 'wrong', 'right'
    timeZero=time.time()
    for k in range(trialsDesired):
        # Get recommended level.  Choose your favorite algorithm.
        tTest=q.quantile()
        #tTest=q.mean()
        #tTest=q.mode()

        tTest=tTest+random.choice([-0.1,0,0.1])

        # Simulate a trial
        timeSplit=time.time(); # omit simulation and printing from reported time/trial.
        response=q.simulate(tTest,tActual)
        print('Trial %3d at %4.1f is %s'%(k+1,tTest,wrongRight[int(response)]))
        timeZero=timeZero+time.time()-timeSplit;

        # Update the pdf
        q.update(tTest,response);

    # Print results of timing.
    print('%.0f ms/trial'%(1000*(time.time()-timeZero)/trialsDesired))

    # Get final estimate.
    t=q.mean()
    sd=q.sd()
    print('Mean threshold estimate is %4.2f +/- %.2f'%(t,sd))
    #t=QuestMode(q);
    #print('Mode threshold estimate is %4.2f'%t)
    print('\nQuest beta analysis. Beta controls the steepness of the Weibull function.\n')
    q.beta_analysis()
    print('Actual parameters of simulated observer:')
    print('logC	beta	gamma')
    print('%5.2f	%4.1f	%5.2f'%(tActual,q.beta,q.gamma))

if __name__ == '__main__':
    demo() # run the demo



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Interface to `EGI Netstation <http://www.egi.com/>`_

This is currently a simple import of
`pynetstation <https://github.com/gaelen/python-egi/>`_
which is now simply called
`egi <https://pypi.python.org/pypi/egi>`_ on pypi.

`egi <https://github.com/gaelen/python-egi/>`_ is included in Standalone
distributions of PsychoPy but you can install it with::

    pip install egi


For examples on usage see the `example_simple` and `example_multi` files on
the `egi github repository <https://github.com/gaelen/python-egi/>`_

For an example see the demos menu of the PsychoPy Coder
For further documentation see the pynetstation website

"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function

from psychopy import logging
try:
    from .egi import *  # pyline: disable=W0614
except ImportError:
    msg = """Failed to import egi (pynetstation). If you're using your own
copy of python (not the Standalone distribution of PsychoPy)
then try installing pynetstation.

See:
    http://code.google.com/p/pynetstation/wiki/Installation

"""
    logging.error(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function

from builtins import str
from os import path
from psychopy.experiment.components import BaseVisualComponent, Param, getInitVals, _translate
from psychopy import logging
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

# the absolute path to the folder containing this path
thisFolder = path.abspath(path.dirname(__file__))
iconFile = path.join(thisFolder, 'brush.png')
tooltip = _translate('Brush: a drawing tool')

# only use _localized values for label values, nothing functional:
_localized.update({'lineWidth': _translate('Brush Size'),
                   'lineColor': _translate('Brush Color'),
                   'lineColorSpace': _translate('Brush Color Space'),
                   'buttonRequired':_translate('Press Button')})

class BrushComponent(BaseVisualComponent):
    """A class for drawing freehand responses"""

    categories = ['Responses']

    def __init__(self, exp, parentName, name='brush',
                 lineColor='$[1,1,1]', lineColorSpace='rgb',
                 lineWidth=1.5, opacity=1,
                 buttonRequired=True,
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim=''):
        super(BrushComponent, self).__init__(
            exp, parentName, name=name,
            startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim)

        self.type = 'Brush'
        self.url = "https://www.psychopy.org/builder/components/brush.html"
        self.exp.requirePsychopyLibs(['visual'])
        self.targets = ['PsychoPy', 'PsychoJS']
        self.order.remove("opacity")  # Move opacity to the end
        self.order += [
            "lineWidth", "lineColor", "lineColorSpace", "opacity"  # Appearance tab
        ]

        # params
        msg = _translate("Fill color of this brush")
        self.params['lineColor'] = Param(
            lineColor, valType='color', inputType="color", allowedTypes=[], categ='Appearance',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat'],
            hint=msg,
            label=_localized['lineColor'])

        msg = _translate("Width of the brush's line (always in pixels and limited to 10px max width)")
        self.params['lineWidth'] = Param(
            lineWidth, valType='num', inputType="spin", allowedTypes=[], categ='Appearance',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat'],
            hint=msg,
            label=_localized['lineWidth'])

        self.params['lineColorSpace'] = self.params['colorSpace']
        del self.params['colorSpace']

        msg = _translate("The line opacity")
        self.params['opacity'].hint=msg

        msg = _translate("Whether a button needs to be pressed to draw (True/False)")
        self.params['buttonRequired'] = Param(
            buttonRequired, valType='bool', inputType="bool", allowedTypes=[], categ='Basic',
            updates='constant',
            allowedUpdates=['constant', 'set every repeat'],
            hint=msg,
            label=_localized['buttonRequired'])

        # Remove BaseVisual params which are not needed
        del self.params['color']  # because color is defined by lineColor
        del self.params['fillColor']
        del self.params['borderColor']
        del self.params['size']  # because size determined by lineWidth
        del self.params['ori']
        del self.params['pos']
        del self.params['units']  # always in pix

    def writeInitCode(self, buff):
        params = getInitVals(self.params)
        code = ("{name} = visual.Brush(win=win, name='{name}',\n"
                "   lineWidth={lineWidth},\n"
                "   lineColor={lineColor},\n"
                "   lineColorSpace={lineColorSpace},\n"
                "   opacity={opacity},\n"
                "   buttonRequired={buttonRequired})").format(name=params['name'],
                                                lineWidth=params['lineWidth'],
                                                lineColor=params['lineColor'],
                                                lineColorSpace=params['lineColorSpace'],
                                                opacity=params['opacity'],
                                                buttonRequired=params['buttonRequired'])
        buff.writeIndentedLines(code)

    def writeInitCodeJS(self, buff):
        # JS code does not use Brush class
        params = getInitVals(self.params)

        code = ("{name} = {{}};\n"
                "get{name} = function() {{\n"
                "  return ( new visual.ShapeStim({{\n"
                "    win: psychoJS.window,\n"
                "    vertices: [[0, 0]],\n"
                "    lineWidth: {lineWidth},\n"
                "    lineColor: new util.Color({lineColor}),\n"
                "    opacity: {opacity},\n"
                "    closeShape: false,\n"
                "    autoLog: false\n"
                "    }}))\n"
                "}}\n\n").format(name=params['name'],
                                 lineWidth=params['lineWidth'],
                                 lineColor=params['lineColor'],
                                 opacity=params['opacity'])

        buff.writeIndentedLines(code)
        # add reset function
        code = ("{name}Reset = {name}.reset = function() {{\n"
                "  if ({name}Shapes.length > 0) {{\n"
                "    for (let shape of {name}Shapes) {{\n"
                "      shape.setAutoDraw(false);\n"
                "    }}\n"
                "  }}\n"
                "  {name}AtStartPoint = false;\n"
                "  {name}Shapes = [];\n"
                "  {name}CurrentShape = -1;\n"
                "}}\n\n").format(name=params['name'])
        buff.writeIndentedLines(code)

        # Define vars for drawing
        code = ("{name}CurrentShape = -1;\n"
                "{name}BrushPos = [];\n"
                "{name}Pointer = new core.Mouse({{win: psychoJS.window}});\n"
                "{name}AtStartPoint = false;\n"
                "{name}Shapes = [];\n").format(name=params['name'])
        buff.writeIndentedLines(code)

    def writeRoutineStartCode(self, buff):
        # Write update code
        super(BrushComponent, self).writeRoutineStartCode(buff)
        # Reset shapes for each trial
        buff.writeIndented("{}.reset()\n".format(self.params['name']))

    def writeRoutineStartCodeJS(self, buff):
        # Write update code
        # super(BrushComponent, self).writeRoutineStartCodeJS(buff)
        # Reset shapes for each trial
        buff.writeIndented("{}Reset();\n".format(self.params['name']))

    def writeFrameCodeJS(self, buff):
        code = ("if ({name}Pointer.getPressed()[0] === 1 && {name}AtStartPoint != true) {{\n"
                "  {name}AtStartPoint = true;\n"
                "  {name}BrushPos = [];\n"
                "  {name}Shapes.push(get{name}());\n"
                "  {name}CurrentShape += 1;\n"
                "  {name}Shapes[{name}CurrentShape].setAutoDraw(true);\n"
                "}}\n"
                "if ({name}Pointer.getPressed()[0] === 1) {{\n"
                "  {name}BrushPos.push({name}Pointer.getPos());\n"
                "  {name}Shapes[{name}CurrentShape].setVertices({name}BrushPos);\n"
                "}} else {{\n"
                "  {name}AtStartPoint = false;\n"
                "}}\n".format(name=self.params['name']))
        buff.writeIndentedLines(code)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division

import os
import pytest
from pandas import DataFrame
from psychopy.visual.window import Window
from psychopy.visual.form import Form
from psychopy.visual.textbox2.textbox2 import TextBox2
from psychopy.visual.slider import Slider
from psychopy import constants
import shutil
from tempfile import mkdtemp
import numpy as np


class Test_Form(object):
    """Test suite for Form component"""

    def setup_class(self):
        # Create temp files for storing items
        self.temp_dir = mkdtemp()
        self.fileName_xlsx = os.path.join(self.temp_dir, 'items.xlsx')
        self.fileName_csv = os.path.join(self.temp_dir, 'items.csv')

        # create some questions
        self.questions = []
        self.genderItem = {"questionText": "What is your gender?",
                           "questionWidth": 0.7,
                           "type": "radio",
                           "responseWidth": 0.3,
                           "options": "Male, Female, Other",
                           "layout": 'vert',
                           "index": 0,
                           "questionColor": "white",
                           "responseColor": "white"
                           }
        self.questions.append(self.genderItem)
        # then a set of ratings
        items = ["running", "cake", "programming"]
        for idx, item in enumerate(items):
            entry = {"questionText": "How much you like {}".format(item),
                     "questionWidth": 0.7,
                     "type": "rating",
                     "responseWidth": 0.3,
                     "options":"Lots, some, Not a lot, Longest Option",
                     "layout": 'horiz',
                     "index": idx+1,
                     "questionColor": "white",
                     "responseColor": "white"
                     }
            self.questions.append(entry)

        self.win = Window(units='height', allowStencil=True, autoLog=False)
        self.survey = Form(self.win, items=self.questions, size=(1.0, 0.3), pos=(0.0, 0.0), autoLog=False)

        # Create datafiles
        df = DataFrame(self.questions)
        df.to_excel(self.fileName_xlsx, index=False)
        df.to_csv(self.fileName_csv, index=False)

    def test_importItems(self):
        wrongFields = [{"a": "What is your gender?",
                      "b": 0.7,
                      "c": "radio",
                      "d": 0.3,
                      "e": "Male, Female, Other",
                      "f": 'vert',
                      "g": "white",
                      "h": "white"
                        }]

        # this doesn't include a itemText or questionText so should get an err
        missingHeader = [{"qText": "What is your gender?",
                      "questionWidth": 0.7,
                      "type": "radio",
                      "responseWidth": 0.3,
                      "options": "Other",
                      "layout": 'vert',
                      "index": 0,
                      "questionColor": "white",
                      "responseColor": "white"}]


        # Check options for list of dicts
        with pytest.raises(ValueError):
            self.survey = Form(self.win, items=missingHeader, size=(1.0, 0.3), pos=(0.0, 0.0), autoLog=False)

        # Check csv
        self.survey = Form(self.win, items=self.fileName_csv,
                           size=(1.0, 0.3), pos=(0.0, 0.0), autoLog=False)
        # Check Excel
        self.survey = Form(self.win, items=self.fileName_xlsx,
                           size=(1.0, 0.3), pos=(0.0, 0.0), randomize=False, autoLog=False)

    def test_set_scroll_speed(self):
        items = 2
        for multipliers in [1,2,3,4]:
            assert self.survey.setScrollSpeed([0] * items, multipliers) == items * multipliers
            assert self.survey.setScrollSpeed([0] * items, multipliers) == items * multipliers
            assert self.survey.setScrollSpeed([0] * items, multipliers) == items * multipliers

    def test_response_text_wrap(self):
        options = ['a', 'b', 'c']
        for size in [.2, .3, .4]:
            item = {"responseWidth": size, "options": options}

    def test_set_questions(self):
        survey = Form(self.win, items=[self.genderItem], size=(1.0, 0.3),
                      pos=(0.0, 0.0), autoLog=False)
        ctrl, h, w = survey._setQuestion(self.genderItem)

        assert type(ctrl) == TextBox2
        assert type(h) in [float, np.float64]
        assert type(w) in [float, np.float64]

    def test_set_response(self):
        survey = Form(self.win, items=[self.genderItem], size=(1.0, 0.3),
                      pos=(0.0, 0.0), autoLog=False)
        ctrl, h, w = survey._setQuestion(self.genderItem)
        sliderStim, respHeight = survey._setResponse(self.genderItem)

        assert type(sliderStim) == Slider
        assert type(respHeight) in [float, np.float64]

    def test_form_size(self):
        assert self.survey.size[0] == (1.0, 0.3)[0]  # width
        assert self.survey.size[1] == (1.0, 0.3)[1]  # height

    def test_aperture_size(self):
        assert self.survey.aperture.size[0] == self.survey.size[0]
        assert self.survey.aperture.size[1] == self.survey.size[1]

    def test_border_limits(self):
        survey = self.survey
        assert survey.leftEdge == survey.pos[0] - survey.size[0]/2.0
        assert survey.rightEdge == survey.pos[0] + survey.size[0]/2.0
        assert survey.topEdge == survey.pos[1] + survey.size[1]/2.0

    def test_text_height(self):
        assert self.survey.textHeight == 0.02

    def test_item_padding(self):
        assert self.survey.itemPadding == 0.05

    def test_form_units(self):
        assert self.survey.units == 'height'

    def test_screen_status(self):
        """Test whether the object is visible"""
        assert self.survey._inRange(self.survey.items[0]['itemCtrl'])

    def test_get_data(self):
        self.survey = Form(self.win, items=self.questions, size=(1.0, 0.3),
                           pos=(0.0, 0.0), autoLog=False)
        data = self.survey.getData()
        Qs = [this['itemText'] for this in data]
        indices = [item['index'] for item in data]
        assert Qs == ['What is your gender?',
                      'How much you like running',
                      'How much you like cake',
                      'How much you like programming',]
        assert all([item['response'] is None for item in data])
        assert all([item['rt'] is None for item in data])
        assert list(indices) == list(range(4))

    def teardown_class(self):
        shutil.rmtree(self.temp_dir)
        self.win.close()


if __name__ == "__main__":
    test = Test_Form()
    test.setup_class()
    test.test_get_data()
    test.teardown_class()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to plotting
"""

from __future__ import absolute_import, print_function


def plotFrameIntervals(intervals):
    """Plot a histogram of the frame intervals.

    Where `intervals` is either a filename to a file, saved by
    Window.saveFrameIntervals, or simply a list (or array) of frame intervals

    """
    from pylab import hist, show, plot

    if type(intervals) == str:
        f = open(intervals, 'r')
        intervals = eval("[%s]" % (f.readline()))
    #    hist(intervals, int(len(intervals)/10))
    plot(intervals)
    show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function
from builtins import super  # provides Py3-style super() using python-future

from os import path
from psychopy.experiment.components import BaseVisualComponent, getInitVals, _translate
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

__author__ = 'Jeremy Gray, Jon Peirce'
# March 2011; builder-component for Yuri Spitsyn's visual.Aperture class
# July 2011: jwp added the code for it to be enabled only when needed

# the absolute path to the folder containing this path
thisFolder = path.abspath(path.dirname(__file__))
iconFile = path.join(thisFolder, 'aperture.png')
tooltip = _translate('Aperture: restrict the drawing of stimuli to a given '
                     'region')


class ApertureComponent(BaseVisualComponent):
    """An event class for using GL stencil to restrict the viewing area to a
    circle or square of a given size and position"""

    def __init__(self, exp, parentName, name='aperture', units='norm',
                 size=1, pos=(0, 0),
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim=''):
        # initialise main parameters
        super(ApertureComponent, self).__init__(
            exp, parentName, name=name, units=units,
            pos=pos, size=size,
            startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim)

        self.type = 'Aperture'
        self.url = "https://www.psychopy.org/builder/components/aperture.html"
        self.order += []

        msg = _translate(
            "How big is the aperture? (a single number for diameter)")
        self.params['size'].hint = msg
        # only localize hints and labels
        self.params['size'].label = _translate("Size")
        self.params['pos'].hint = _translate("Where is the aperture centred?")

        # Remove BaseVisual params which are not needed
        del self.params['ori']
        del self.params['color']
        del self.params['colorSpace']
        del self.params['fillColor']
        del self.params['borderColor']
        del self.params['opacity']

    def writeInitCode(self, buff):
        # do writing of init
        inits = getInitVals(self.params)

        # do we need units code?
        if self.params['units'].val == 'from exp settings':
            unitsStr = ""
        else:
            unitsStr = f"units={inits['units']},"

        code = (f"{inits['name']} = visual.Aperture(\n"
                f"    win=win, name='{inits['name']}',\n"
                f"    {unitsStr} size={inits['size']}, pos={inits['pos']})\n"
                f"{inits['name']}.disable()  # disable until its actually used\n")
        buff.writeIndentedLines(code)

    def writeFrameCode(self, buff):
        """Only activate the aperture for the required frames
        """
        params = self.params
        code = (f"\n"
                f"# *{params['name']}* updates\n")
        buff.writeIndented(code)
        # writes an if statement to determine whether to draw etc
        self.writeStartTestCode(buff)
        buff.writeIndented("%(name)s.enabled = True\n" % self.params)
        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)
        if self.params['stopVal'].val not in ['', None, -1, 'None']:
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCode(buff)
            buff.writeIndented("%(name)s.enabled = False\n" % self.params)
            # to get out of the if statement
            buff.setIndentLevel(-2, relative=True)
        # set parameters that need updating every frame
        # do any params need updating? (this method inherited from _base)
        if self.checkNeedToUpdate('set every frame'):
            code = ("if %(name)s.status == STARTED:  # only update if being  drawn\n")
            buff.writeIndented(code % self.params)

            buff.setIndentLevel(+1, relative=True)  # to enter the if block
            self.writeParamUpdates(buff, 'set every frame')
            buff.setIndentLevel(-1, relative=True)  # to exit the if block

    def writeRoutineEndCode(self, buff):
        msg = "%(name)s.enabled = False  # just in case it was left enabled\n"
        buff.writeIndented(msg % self.params)

        # get parent to write code too (e.g. store onset/offset times)
        super().writeRoutineEndCode(buff)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import, print_function

from builtins import next
from past.builtins import basestring
from builtins import object
import numbers  # numbers.Integral is like (int, long) but supports Py3
from tables import *
import os
from collections import namedtuple
import json

from ..errors import print2err

from pkg_resources import parse_version
import tables
if parse_version(tables.__version__) < parse_version('3'):
    from tables import openFile as open_file
    walk_groups = "walkGroups"
    list_nodes = "listNodes"
    get_node = "getNode"
    read_where = "readWhere"
else:
    from tables import open_file
    walk_groups = "walk_groups"
    list_nodes = "list_nodes"
    get_node = "get_node"
    read_where = "read_where"


_hubFiles = []

def openHubFile(filepath, filename, mode):
    """
    Open an HDF5 DataStore file and register it so that it is closed even on interpreter crash.
    """
    global _hubFiles
    hubFile = open_file(os.path.join(filepath, filename), mode)
    _hubFiles.append(hubFile)
    return hubFile


def displayDataFileSelectionDialog(starting_dir=None):
    """Shows a FileDialog and lets you select a .hdf5 file to open for
    processing."""
    from psychopy.gui.qtgui import fileOpenDlg

    filePath = fileOpenDlg(tryFilePath=starting_dir, 
                           prompt = "Select a ioHub HDF5 File",
                           allowed='HDF5 Files (*.hdf5)')
    

    if filePath is None:
        return None

    return filePath


def displayEventTableSelectionDialog(
        title,
        list_label,
        list_values,
        default=u'Select'):
    from psychopy import gui
    if default not in list_values:
        list_values.insert(0, default)
    else:
        list_values.remove(list_values)
        list_values.insert(0, default)

    selection_dict = dict(list_label=list_values)
    dlg_info = dict(selection_dict)
    infoDlg = gui.DlgFromDict(dictionary=dlg_info, title=title)
    if not infoDlg.OK:
        return None

    while list(dlg_info.values())[0] == default and infoDlg.OK:
            dlg_info=dict(selection_dict)
            infoDlg = gui.DlgFromDict(dictionary=dlg_info, title=title)

    if not infoDlg.OK:
        return None

    return list(dlg_info.values())[0]
########### Experiment / Experiment Session Based Data Access #################


class ExperimentDataAccessUtility(object):
    """The ExperimentDataAccessUtility  provides a simple, high level, way to
    access data saved in an ioHub DataStore HDF5 file. Data access is done by
    providing information at an experiment and session level, as well as
    specifying the ioHub Event types you want to retieve data for.

    An instance of the ExperimentDataAccessUtility class is created by providing
    the location and name of the file to read, as well as any session code
    filtering you want applied to the retieved datasets.

    Args:
        hdfFilePath (str): The path of the directory the DataStore HDF5 file is in.

        hdfFileName (str): The name of the DataStore HDF5 file.

        experimentCode (str): If multi-experiment support is enabled for the DataStore file, this arguement can be used to specify what experiment data to load based on the experiment_code given. NOTE: Multi-experiment data file support is not well tested and should not be used at this point.

        sessionCodes (str or list): The experiment session code to filter data by. If a list of codes is given, then all codes in the list will be used.

    Returns:
        object: the created instance of the ExperimentDataAccessUtility, ready to get your data!

    """

    def __init__(
            self,
            hdfFilePath,
            hdfFileName,
            experimentCode=None,
            sessionCodes=[],
            mode='r'):
        """An instance of the ExperimentDataAccessUtility class is created by
        providing the location and name of the file to read, as well as any
        session code filtering you want applied to the retieved datasets.

        Args:
            hdfFilePath (str): The path of the directory the DataStore HDF5 file is in.

            hdfFileName (str): The name of the DataStore HDF5 file.

            experimentCode (str): If multi-experiment support is enabled for the DataStore file, this arguement can be used to specify what experiment data to load based on the experiment_code given. NOTE: Multi-experiment data file support is not well tested and should not be used at this point.

            sessionCodes (str or list): The experiment session code to filter data by. If a list of codes is given, then all codes in the list will be used.

        Returns:
            object: the created instance of the ExperimentDataAccessUtility, ready to get your data!

        """
        self.hdfFilePath = hdfFilePath
        self.hdfFileName = hdfFileName
        self.mode = mode
        self.hdfFile = None

        self._experimentCode = experimentCode
        self._sessionCodes = sessionCodes
        self._lastWhereClause = None

        try:
            self.hdfFile = openHubFile(hdfFilePath, hdfFileName, mode)
        except Exception as e:
            print(e)
            raise ExperimentDataAccessException(e)

        self.getExperimentMetaData()

    def printTableStructure(self,tableName):
        """Print to stdout the current structure and content statistics of the
        specified DataStore table. To print out the complete structure of the
        DataStore file, including the name of all available tables, see the
        printHubFileStructure method.

        Args:
            tableName (str): The DataStore table name to print metadata information out for.

        """
        if self.hdfFile:
            hubFile = self.hdfFile
            for group in getattr(hubFile, walk_groups)("/"):
                for table in getattr(hubFile, list_nodes)(group, classname='Table'):
                    if table.name == tableName:
                        print('------------------')
                        print('Path:', table)
                        print('Table name:', table.name)
                        print('Number of rows in table:', table.nrows)
                        print('Number of cols in table:', len(table.colnames))
                        print('Attribute name := type, shape:')
                        for name in table.colnames:
                            print('\t', name, ':= %s, %s' % (table.coldtypes[name], table.coldtypes[name].shape))
                        print('------------------')
                        return

    def printHubFileStructure(self):
        """Print to stdout the current global structure of the loaded DataStore
        File."""
        if self.hdfFile:
            print(self.hdfFile)

    def getExperimentMetaData(self):
        """Returns the the metadata for the experiment the datStore file is
        for.

        **Docstr TBC.**

        """
        if self.hdfFile:
            expcols = self.hdfFile.root.data_collection.experiment_meta_data.colnames
            if 'sessions' not in expcols:
                expcols.append('sessions')
            ExperimentMetaDataInstance = namedtuple(
                'ExperimentMetaDataInstance', expcols)
            experiments=[]
            for e in self.hdfFile.root.data_collection.experiment_meta_data:
                self._experimentID = e['experiment_id']
                a_exp = list(e[:])
                a_exp.append(self.getSessionMetaData())
                experiments.append(ExperimentMetaDataInstance(*a_exp))
            return experiments

    def getSessionMetaData(self, sessions=None):
        """
        Returns the the metadata associated with the experiment session codes in use.

        **Docstr TBC.**

        """
        if self.hdfFile:
            if sessions is None:
                sessions = []

            sessionCodes = self._sessionCodes
            sesscols = self.hdfFile.root.data_collection.session_meta_data.colnames
            SessionMetaDataInstance = namedtuple('SessionMetaDataInstance', sesscols)
            for r in self.hdfFile.root.data_collection.session_meta_data:
                if (len(sessionCodes) == 0 or r['code'] in sessionCodes) and r[
                        'experiment_id'] == self._experimentID:
                    rcpy=list(r[:])
                    rcpy[-1]=json.loads(rcpy[-1])
                    sessions.append(SessionMetaDataInstance(*rcpy))
            return sessions

    def getTableForPath(self, path):
        """
        Given a valid table path within the DataStore file, return the accociated table.
        """
        getattr(self.hdfFile, get_node)(path)

    def getEventTable(self, event_type):
        """
        Returns the DataStore table that contains events of the specified type.

        **Docstr TBC.**

        """
        if self.hdfFile:
            klassTables = self.hdfFile.root.class_table_mapping
            event_column = None
            event_value = None

            if isinstance(event_type, basestring):
                if event_type.find('Event') >= 0:
                    event_column = 'class_name'
                    event_value = event_type
                else:
                    event_value = ''
                    tokens = event_type.split('_')
                    for t in tokens:
                        event_value += t[0].upper()+t[1:].lower()
                    event_value = event_type+'Event'
            elif isinstance(event_type, numbers.Integral):
                event_column = 'class_id'
                event_value = event_type
            else:
                print2err(
                    'getEventTable error: event_type arguement must be a string or and int')
                return None

            result = []
            where_cls = '(%s == b"%s") & (class_type_id == 1)'%(event_column, event_value)
            for row in klassTables.where(where_cls):
                result.append(row.fetch_all_fields())

            if len(result) == 0:
                    return None
                    
            if len(result)!= 1:
                print2err(
                    'event_type_id passed to getEventAttribute can only return one row from CLASS_MAPPINGS: ',
                    len(result))
                return None
            tablePathString = result[0][3]
            if isinstance(tablePathString, bytes):
                tablePathString = tablePathString.decode('utf-8')         
            return getattr(self.hdfFile, get_node)(tablePathString)
        return None

    def getEventMappingInformation(self):
        """Returns details on how ioHub Event Types are mapped to tables within
        the given DataStore file."""
        if self.hdfFile:
            eventMappings=dict()
            class_2_table=self.hdfFile.root.class_table_mapping
            EventTableMapping = namedtuple(
                'EventTableMapping',
                self.hdfFile.root.class_table_mapping.colnames)
            for row in class_2_table[:]:
                eventMappings[row['class_id']] = EventTableMapping(*row)
            return eventMappings
        return None

    def getEventsByType(self, condition_str = None):
        """Returns a dict of all event tables within the DataStore file that
        have at least one event instance saved.

        Keys are Event Type constants, as specified by
        iohub.EventConstants. Each value is a row iterator for events of
        that type.

        """
        eventTableMappings = self.getEventMappingInformation()
        if eventTableMappings:
            events_by_type = dict()
            getNode = getattr(self.hdfFile, get_node)
            for event_type_id, event_mapping_info in eventTableMappings.items():
                try:
                    cond = '(type == %d)' % (event_type_id)
                    if condition_str:
                        cond += ' & ' + condition_str
                    et_path = event_mapping_info.table_path
                    if isinstance(et_path, bytes):
                        et_path = et_path.decode('utf-8')
                    events_by_type[event_type_id] = next(getNode(et_path).where(cond))
                except StopIteration:
                    pass
            return events_by_type
        return None

    def getConditionVariablesTable(self):
        """
        **Docstr TBC.**
        """
        cv_group = self.hdfFile.root.data_collection.condition_variables
        ecv = 'EXP_CV_%d' % (self._experimentID,)
        if ecv in cv_group._v_leaves:
            return cv_group._v_leaves[ecv]
        return None

    def getConditionVariableNames(self):
        """
        **Docstr TBC.**
        """
        cv_group = self.hdfFile.root.data_collection.condition_variables
        ecv = "EXP_CV_%d" % (self._experimentID,)
        if ecv in cv_group._v_leaves:
            ecvTable = cv_group._v_leaves[ecv]
            return ecvTable.colnames
        return None

    def getConditionVariables(self, filter=None):
        """
        **Docstr TBC.**
        """
        if filter is None:
            session_ids = []
            for s in self.getExperimentMetaData()[0].sessions:
                session_ids.append(s.session_id)
            filter = dict(session_id=(' in ', session_ids))

        ConditionSetInstance = None

        for conditionVarName, conditionVarComparitor in filter.items():
            avComparison, value = conditionVarComparitor

            cv_group = self.hdfFile.root.data_collection.condition_variables
            cvrows = []
            ecv = "EXP_CV_%d" % (self._experimentID,)
            if ecv in cv_group._v_leaves:
                ecvTable = cv_group._v_leaves[ecv]

                if ConditionSetInstance is None:
                    colnam = ecvTable.colnames
                    ConditionSetInstance = namedtuple('ConditionSetInstance', colnam)

                cvrows.extend(
                    [
                        ConditionSetInstance(
                            *
                            r[:]) for r in ecvTable if all(
                            [
                                eval(
                                    '{0} {1} {2}'.format(
                                        r[conditionVarName],
                                        conditionVarComparitor[0],
                                        conditionVarComparitor[1])) for conditionVarName,
                                conditionVarComparitor in filter.items()])])
        return cvrows

    def getValuesForVariables(self, cv, value, cvNames):
        """
        **Docstr TBC.**
        """

        if isinstance(value, (list, tuple)):
            resolvedValues = []
            for v in value:
                if isinstance(value, basestring) and value.startswith(
                        '@') and value.endswith('@'):
                    value=value[1:-1]
                    if value in cvNames:
                        resolvedValues.append(getattr(cv, v))
                    else:
                        raise ExperimentDataAccessException(
                            'getEventAttributeValues: {0} is not a valid attribute name in {1}'.format(
                                v, cvNames))
                elif isinstance(value, basestring):
                    resolvedValues.append(value)
            return resolvedValues
        elif isinstance(value, basestring) and value.startswith('@') and value.endswith('@'):
            value = value[1:-1]
            if value in cvNames:
                return getattr(cv, value)
            else:
                raise ExperimentDataAccessException(
                    'getEventAttributeValues: {0} is not a valid attribute name in {1}'.format(
                        value, cvNames))
        else:
            raise ExperimentDataAccessException(
                'Unhandled value type !: {0} is not a valid type for value {1}'.format(
                    type(value), value))

    def getEventAttributeValues(
            self,
            event_type_id,
            event_attribute_names,
            filter_id=None,
            conditionVariablesFilter=None,
            startConditions=None,
            endConditions=None):
        """
        **Docstr TBC.**

        Args:
            event_type_id
            event_attribute_names
            conditionVariablesFilter
            startConditions
            endConditions

        Returns:
            Values for the specified event type and event attribute columns which match the provided experiment condition variable filter, starting condition filer, and ending condition filter criteria.
        """
        if self.hdfFile:
            klassTables = self.hdfFile.root.class_table_mapping

            deviceEventTable = None

            result = [
                row.fetch_all_fields() for row in klassTables.where(
                    '(class_id == %d) & (class_type_id == 1)' %
                    (event_type_id))]
            if len(result) != 1:
                raise ExperimentDataAccessException("event_type_id passed to getEventAttribute should only return one row from CLASS_MAPPINGS.")
            tablePathString = result[0][3]
            deviceEventTable = getattr(self.hdfFile, get_node)(tablePathString)

            for ename in event_attribute_names:
                if ename not in deviceEventTable.colnames:
                    raise ExperimentDataAccessException(
                        'getEventAttribute: %s does not have a column named %s' %
                        (deviceEventTable.title, event_attribute_names))

            resultSetList = []

            csier = list(event_attribute_names)
            csier.append('query_string')
            csier.append('condition_set')
            EventAttributeResults = namedtuple('EventAttributeResults', csier)

            if deviceEventTable is not None:
                if not isinstance(event_attribute_names, (list, tuple)):
                    event_attribute_names = [event_attribute_names, ]

                filteredConditionVariableList = None
                if conditionVariablesFilter is None:
                    filteredConditionVariableList= self.getConditionVariables()
                else:
                    filteredConditionVariableList = self.getConditionVariables(
                        conditionVariablesFilter)

                cvNames = self.getConditionVariableNames()

                # no further where clause building needed; get reseults and
                # return
                if startConditions is None and endConditions is None:
                    for cv in filteredConditionVariableList:

                        wclause = '( experiment_id == {0} ) & ( session_id == {1} )'.format(
                            self._experimentID, cv.session_id)

                        wclause += ' & ( type == {0} ) '.format(event_type_id)

                        if filter_id is not None:
                            wclause += '& ( filter_id == {0} ) '.format(
                                filter_id)

                        resultSetList.append([])

                        for ename in event_attribute_names:
                            resultSetList[-1].append(getattr(deviceEventTable, read_where)(wclause, field=ename))
                        resultSetList[-1].append(wclause)
                        resultSetList[-1].append(cv)

                        eventAttributeResults = EventAttributeResults(
                            *resultSetList[-1])
                        resultSetList[-1]=eventAttributeResults

                    return resultSetList

                #start or end conditions exist....
                for cv in filteredConditionVariableList:
                    resultSetList.append([])

                    wclause = '( experiment_id == {0} ) & ( session_id == {1} )'.format(
                        self._experimentID, cv.session_id)

                    wclause += ' & ( type == {0} ) '.format(event_type_id)

                    if filter_id is not None:
                        wclause += '& ( filter_id == {0} ) '.format(filter_id)

                    # start Conditions need to be added to where clause
                    if startConditions is not None:
                        wclause += '& ('
                        for conditionAttributeName, conditionAttributeComparitor in startConditions.items():
                            avComparison,value=conditionAttributeComparitor
                            value = self.getValuesForVariables(
                                cv, value, cvNames)
                            wclause += ' ( {0} {1} {2} ) & '.format(
                                conditionAttributeName, avComparison, value)
                        wclause=wclause[:-3]
                        wclause += ' ) '

                    # end Conditions need to be added to where clause
                    if endConditions is not None:
                        wclause += ' & ('
                        for conditionAttributeName, conditionAttributeComparitor in endConditions.items():
                            avComparison,value=conditionAttributeComparitor
                            value = self.getValuesForVariables(
                                cv, value, cvNames)
                            wclause += ' ( {0} {1} {2} ) & '.format(
                                conditionAttributeName, avComparison, value)
                        wclause=wclause[:-3]
                        wclause += ' ) '

                    for ename in event_attribute_names:
                        resultSetList[-1].append(getattr(deviceEventTable, read_where)(wclause, field=ename))
                    resultSetList[-1].append(wclause)
                    resultSetList[-1].append(cv)

                    eventAttributeResults = EventAttributeResults(
                        *resultSetList[-1])
                    resultSetList[-1]=eventAttributeResults

                return resultSetList

            return None

    def getEventIterator(self, event_type):
        """
        **Docstr TBC.**

        Args:
            event_type

        Returns:
            (iterator): An iterator providing access to each matching event  as a numpy recarray.
        """
        return self.getEventTable(event_type).iterrows()

    def close(self):
        """Close the ExperimentDataAccessUtility and associated DataStore
        File."""
        global _hubFiles
        if self.hdfFile in _hubFiles:
            _hubFiles.remove(self.hdfFile)
        self.hdfFile.close()

        self.experimentCodes = None
        self.hdfFilePath = None
        self.hdfFileName = None
        self.mode = None
        self.hdfFile = None

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass


class ExperimentDataAccessException(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Stimulus class for drawing radial stimuli.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from builtins import str

# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+
import pyglet

from ..colors import Color

pyglet.options['debug_gl'] = False
import ctypes
GL = pyglet.gl

import psychopy  # so we can get the __path__
from psychopy import logging

# tools must only be imported *after* event or MovieStim breaks on win32
# (JWP has no idea why!)
from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import attributeSetter, setAttribute
from psychopy.visual.grating import GratingStim

try:
    from PIL import Image
except ImportError:
    from . import Image

import numpy
from numpy import pi


class RadialStim(GratingStim):
    """Stimulus object for drawing radial stimuli.

    Examples: annulus, rotating wedge, checkerboard.

    Ideal for fMRI retinotopy stimuli!

    Many of the capabilities are built on top of the GratingStim.

    This stimulus is still relatively new and I'm finding occasional glitches.
    It also takes longer to draw than a typical GratingStim, so not
    recommended for tasks where high frame rates are needed.
    """

    def __init__(self,
                 win,
                 tex="sqrXsqr",
                 mask="none",
                 units="",
                 pos=(0.0, 0.0),
                 size=(1.0, 1.0),
                 radialCycles=3,
                 angularCycles=4,
                 radialPhase=0,
                 angularPhase=0,
                 ori=0.0,
                 texRes=64,
                 angularRes=100,
                 visibleWedge=(0, 360),
                 rgb=None,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 dkl=None,
                 lms=None,
                 contrast=1.0,
                 opacity=1.0,
                 depth=0,
                 rgbPedestal=(0.0, 0.0, 0.0),
                 interpolate=False,
                 name=None,
                 autoLog=None,
                 maskParams=None):
        """ """  # Empty docstring on __init__
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        self._initParams.remove('self')

        super(RadialStim, self).__init__(win, units=units, name=name,
                                         autoLog=False)  # start off false

        # use shaders if available by default, this is a good thing
        self.useShaders = win._haveShaders

        # UGLY HACK again. (See same section in GratingStim for ideas)
        self.__dict__['contrast'] = 1
        self.__dict__['size'] = 1
        self.__dict__['sf'] = 1
        self.__dict__['tex'] = tex

        # initialise textures for stimulus
        self._texID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._texID))
        self._maskID = GL.GLuint()
        GL.glGenTextures(1, ctypes.byref(self._maskID))
        self.__dict__['maskParams'] = maskParams
        self.maskRadialPhase = 0
        self.texRes = texRes  # must be power of 2
        self.interpolate = interpolate
        self.rgbPedestal = val2array(rgbPedestal, False, length=3)

        # these are defined for GratingStim but can only cause confusion here
        self.setSF = None
        self.setPhase = None

        self.colorSpace = colorSpace
        if rgb != None:
            logging.warning("Use of rgb arguments to stimuli are deprecated."
                            " Please use color and colorSpace args instead")
            self.color = Color(rgb, colorSpace='rgb')
        elif dkl != None:
            logging.warning("Use of dkl arguments to stimuli are deprecated. "
                            "Please use color and colorSpace args instead")
            self.color = Color(dkl, colorSpace='dkl')
        elif lms != None:
            logging.warning("Use of lms arguments to stimuli are deprecated."
                            " Please use color and colorSpace args instead")
            self.color = Color(lms, colorSpace='lms')
        else:
            self.color = color

        self.ori = float(ori)
        self.__dict__['angularRes'] = angularRes
        self.__dict__['radialPhase'] = radialPhase
        self.__dict__['radialCycles'] = radialCycles
        self.__dict__['visibleWedge'] = numpy.array(visibleWedge)
        self.__dict__['angularCycles'] = angularCycles
        self.__dict__['angularPhase'] = angularPhase
        self.pos = numpy.array(pos, float)
        self.depth = depth
        self.__dict__['sf'] = 1
        self.size = val2array(size, False)

        self.tex = tex
        self.mask = mask
        self.contrast = float(contrast)
        self.opacity = float(opacity)

        #
        self._updateEverything()

        # set autoLog now that params have been initialised
        wantLog = autoLog is None and self.win.autoLog
        self.__dict__['autoLog'] = autoLog or wantLog
        if self.autoLog:
            logging.exp("Created %s = %s" % (self.name, str(self)))

    @attributeSetter
    def mask(self, value):
        """The alpha mask that forms the shape of the resulting image.

        Value should be one of:

            + 'circle', 'gauss', 'raisedCos', **None** (resets to default)
            + or the name of an image file (most formats supported)
            + or a numpy array (1xN) ranging -1:1

        Note that the mask for `RadialStim` is somewhat different to the
        mask for :class:`ImageStim`. For `RadialStim` it is a 1D array
        specifying the luminance profile extending outwards from the
        center of the stimulus, rather than a 2D array
        """
        # todo: fromFile is not used
        fromFile = 0
        self.__dict__['mask'] = value
        res = self.texRes  # resolution of texture - 128 is bearable
        step = 1.0/res
        rad = numpy.arange(0, 1 + step, step)
        if type(self.mask) == numpy.ndarray:
            # handle a numpy array
            intensity = 255 * self.mask.astype(float)
            res = len(intensity)
        elif type(self.mask) == list:
            # handle a numpy array
            intensity = 255 * numpy.array(self.mask, float)
            res = len(intensity)
        elif self.mask == "circle":
            intensity = 255.0 * (rad <= 1)
        elif self.mask == "gauss":
            # Set SD if specified
            if self.maskParams is None:
                sigma = 1.0/3
            else:
                sigma = 1.0/self.maskParams['sd']
            # 3sd.s by the edge of the stimulus
            intensity = 255.0 * numpy.exp(-rad**2.0/(2.0 * sigma**2.0))
        elif self.mask == "radRamp":  # a radial ramp
            intensity = 255.0 - 255.0 * rad
            # half wave rectify:
            intensity = numpy.where(rad < 1, intensity, 0)
        elif self.mask in [None, "none", "None"]:
            res = 4
            intensity = 255.0 * numpy.ones(res, float)
        else:  # might be a filename of a tiff
            try:
                im = Image.open(self.mask)
                im = im.transpose(Image.FLIP_TOP_BOTTOM)
                im = im.resize([max(im.size), max(im.size)],
                               Image.BILINEAR)  # make it square
            except IOError as details:
                msg = "couldn't load mask...%s: %s"
                logging.error(msg % (value, details))
                return
            res = im.size[0]
            im = im.convert("L")  # force to intensity (in case it was rgb)
            intensity = numpy.asarray(im)
            fromFile = 1

        data = intensity.astype(numpy.uint8)
        mask = data.tostring()  # serialise

        # do the openGL binding
        if self.interpolate:
            smoothing = GL.GL_LINEAR
        else:
            smoothing = GL.GL_NEAREST
        GL.glBindTexture(GL.GL_TEXTURE_1D, self._maskID)
        GL.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_ALPHA,
                        res, 0, GL.GL_ALPHA, GL.GL_UNSIGNED_BYTE, mask)
        # makes the texture map wrap (this is actually default anyway)
        GL.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_WRAP_S,
                           GL.GL_REPEAT)
        # linear smoothing if texture is stretched
        GL.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MAG_FILTER,
                           smoothing)
        GL.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER,
                           smoothing)
        GL.glTexEnvi(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE,
                     GL.GL_MODULATE)
        GL.glEnable(GL.GL_TEXTURE_1D)

        self._needUpdate = True

    def setMask(self, value, log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'mask', value, log)

    def _setRadialAtribute(self, attr, value):
        """Internal helper function to reduce redundancy
        """

        self.__dict__[attr] = value  # avoid recursing the attributeSetter
        self._updateTextureCoords()
        self._needUpdate = True

    @attributeSetter
    def angularCycles(self, value):
        """Float (but Int is prettiest). Set the number of cycles going
        around the stimulus. i.e. it controls the number of 'spokes'.

        :ref:`Operations <attrib-operations>` supported.
        """
        self._setRadialAtribute('angularCycles', value)

    def setAngularCycles(self, value, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'angularCycles', value, log,
                     operation)  # calls the attributeSetter

    @attributeSetter
    def radialCycles(self, value):
        """Float (but Int is prettiest). Set the number of texture cycles
        from centre to periphery, i.e. it controls the number of 'rings'.

        :ref:`Operations <attrib-operations>` supported.
        """
        self._setRadialAtribute('radialCycles', value)

    def setRadialCycles(self, value, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'radialCycles', value, log,
                     operation)  # calls the attributeSetter

    @attributeSetter
    def angularPhase(self, value):
        """Float. Set the angular phase (like orientation) of the texture
        (wraps 0-1).

        This is akin to setting the orientation of the texture around the
        stimulus in radians. If possible, it is more efficient to rotate the
        stimulus using its `ori` setting instead.

        :ref:`Operations <attrib-operations>` supported.
        """
        self._setRadialAtribute('angularPhase', value)

    def setAngularPhase(self, value, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'angularPhase', value, log,
                     operation)  # calls the attributeSetter

    @attributeSetter
    def radialPhase(self, value):
        """Float. Set the radial phase of the texture (wraps 0-1). This is the
        phase of the texture from the centre to the perimeter of the stimulus
        (in radians). Can be used to drift concentric rings out/inwards.

        :ref:`Operations <attrib-operations>` supported.
        """
        self._setRadialAtribute('radialPhase', value)

    def setRadialPhase(self, value, operation='', log=None):
        """Usually you can use 'stim.attribute = value' syntax instead,
        but use this method if you need to suppress the log message
        """
        setAttribute(self, 'radialPhase', value, log,
                     operation)  # calls the attributeSetter

    def _updateEverything(self):
        """Internal helper function for angularRes and visibleWedge (and init)
        """
        self._triangleWidth = pi * 2 / self.angularRes
        self._angles = numpy.arange(0, pi * 2, self._triangleWidth,
                                    dtype='float64')
        # which vertices are visible?
        # first edge of wedge:
        visW = self.visibleWedge
        self._visible = (self._angles >= visW[0] * pi / 180)
        # second edge of wedge:
        edge2 = (self._angles + self._triangleWidth) * (180/pi) > visW[1]
        self._visible[edge2] = False
        self._nVisible = numpy.sum(self._visible) * 3

        self._updateTextureCoords()
        self._updateMaskCoords()
        self._updateVerticesBase()
        self._updateVertices()  # is this necessary? Works fine without...
        if not self.useShaders:
            # generate a displaylist ID
            self._listID = GL.glGenLists(1)
            self._updateList()  # ie refresh display list

    @attributeSetter
    def angularRes(self, value):
        """The number of triangles used to make the sti.

         :ref:`Operations <attrib-operations>` supported."""
        self.__dict__['angularRes'] = value
        self._updateEverything()

    @attributeSetter
    def visibleWedge(self, value):
        """tuple (start, end) in degrees. Determines visible range.

        (0, 360) is full visibility.

        :ref:`Operations <attrib-operations>` supported.
        """
        self.__dict__['visibleWedge'] = numpy.array(value)
        self._updateEverything()

    def draw(self, win=None):
        """Draw the stimulus in its relevant window. You must call
        this method after every `win.flip()` if you want the
        stimulus to appear on that frame and then update the screen
        again.

        If `win` is specified then override the normal window of this
        stimulus.
        """
        if win is None:
            win = self.win
        self._selectWindow(win)

        # do scaling
        GL.glPushMatrix()  # push before the list, pop after
        # scale the viewport to the appropriate size
        self.win.setScale('pix')
        if self.useShaders:
            # setup color
            GL.glColor4f(*self._foreColor.render('rgba1'))

            # assign vertex array
            GL.glVertexPointer(2, GL.GL_DOUBLE, 0, self.verticesPix.ctypes)

            # then bind main texture
            GL.glActiveTexture(GL.GL_TEXTURE0)
            GL.glBindTexture(GL.GL_TEXTURE_2D, self._texID)
            GL.glEnable(GL.GL_TEXTURE_2D)
            # and mask
            GL.glActiveTexture(GL.GL_TEXTURE1)
            GL.glBindTexture(GL.GL_TEXTURE_1D, self._maskID)
            GL.glDisable(GL.GL_TEXTURE_2D)
            GL.glEnable(GL.GL_TEXTURE_1D)

            # setup the shaderprogram
            prog = self.win._progSignedTexMask1D
            GL.glUseProgram(prog)
            # set the texture to be texture unit 0
            GL.glUniform1i(GL.glGetUniformLocation(prog, b"texture"), 0)
            # mask is texture unit 1
            GL.glUniform1i(GL.glGetUniformLocation(prog, b"mask"), 1)

            # set pointers to visible textures
            GL.glClientActiveTexture(GL.GL_TEXTURE0)
            GL.glTexCoordPointer(2, GL.GL_DOUBLE, 0,
                                 self._visibleTexture.ctypes)
            GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)

            # mask
            GL.glClientActiveTexture(GL.GL_TEXTURE1)
            GL.glTexCoordPointer(1, GL.GL_DOUBLE, 0,
                                 self._visibleMask.ctypes)
            GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)

            # do the drawing
            GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
            GL.glDrawArrays(GL.GL_TRIANGLES, 0, self._nVisible)

            # unbind the textures
            GL.glClientActiveTexture(GL.GL_TEXTURE1)
            GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
            # main texture
            GL.glClientActiveTexture(GL.GL_TEXTURE0)
            GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
            GL.glDisable(GL.GL_TEXTURE_2D)
            # disable set states
            GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
            GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY)

            GL.glUseProgram(0)
        else:
            # the list does the texture mapping
            if self._needTextureUpdate:
                self.setTex(value=self.tex, log=False)
            if self._needUpdate:
                self._updateList()
            GL.glCallList(self._listID)

        # return the view to previous state
        GL.glPopMatrix()

    def _updateVerticesBase(self):
        """Update the base vertices if angular resolution changes.

        These will be multiplied by the size and rotation matrix before
        rendering.
        """
        # triangles = [trisX100, verticesX3, xyX2]
        vertsBase = numpy.zeros([self.angularRes, 3, 2])
        # x position of 1st outer vertex
        vertsBase[:, 1, 0] = numpy.sin(self._angles)
        # y position of 1st outer vertex
        vertsBase[:, 1, 1] = numpy.cos(self._angles)
        # x position of 2nd outer vertex
        vertsBase[:, 2, 0] = numpy.sin(self._angles + self._triangleWidth)
        # y position of 2nd outer vertex
        vertsBase[:, 2, 1] = numpy.cos(self._angles + self._triangleWidth)
        vertsBase /= 2.0  # size should be 1.0, so radius should be 0.5
        vertsBase = vertsBase[self._visible, :, :]
        self._verticesBase = vertsBase.reshape(self._nVisible, 2)

    def _updateTextureCoords(self):
        """calculate texture coordinates if angularCycles or Phase change
        """
        pi2 = 2 * pi
        self._textureCoords = numpy.zeros([self.angularRes, 3, 2])
        # x position of inner vertex
        self._textureCoords[:, 0, 0] = (
            (self._angles + self._triangleWidth/2) *
            self.angularCycles / pi2 + self.angularPhase)
        # y position of inner vertex
        self._textureCoords[:, 0, 1] = 0.25 - self.radialPhase

        # x position of 1st outer vertex
        self._textureCoords[:, 1, 0] = (
            self._angles * self.angularCycles / pi2 + self.angularPhase)
        # y position of 1st outer vertex
        self._textureCoords[:, 1, 1] = (
            0.25 + self.radialCycles - self.radialPhase)

        # x position of 2nd outer vertex
        self._textureCoords[:, 2, 0] = (
            (self._angles + self._triangleWidth) *
            self.angularCycles / pi2 + self.angularPhase)
        # y position of 2nd outer vertex
        self._textureCoords[:, 2, 1] = (
            0.25 + self.radialCycles - self.radialPhase)
        self._visibleTexture = self._textureCoords[
            self._visible, :, :].reshape(self._nVisible, 2)

    def _updateMaskCoords(self):
        """calculate mask coords
        """
        self._maskCoords = numpy.zeros(
            [self.angularRes, 3]) + self.maskRadialPhase
        # all outer points have mask value of 1
        self._maskCoords[:, 1:] = 1 + self.maskRadialPhase
        self._visibleMask = self._maskCoords[self._visible, :]

    def _updateListShaders(self):
        """The user shouldn't need this method since it gets called
        after every call to .set() Basically it updates the OpenGL
        representation of your stimulus if some parameter of the
        stimulus changes. Call it if you change a property manually
        rather than using the .set() command
        """
        self._needUpdate = False
        GL.glNewList(self._listID, GL.GL_COMPILE)

        # assign vertex array
        arrPointer = self.verticesPix.ctypes.data_as(
            ctypes.POINTER(ctypes.c_float))
        GL.glVertexPointer(2, GL.GL_FLOAT, 0, arrPointer)

        # setup the shaderprogram
        GL.glUseProgram(self.win._progSignedTexMask1D)
        # set the texture to be texture unit 0
        GL.glUniform1i(GL.glGetUniformLocation(
            self.win._progSignedTexMask1D, b"texture"), 0)
        GL.glUniform1i(GL.glGetUniformLocation(
            self.win._progSignedTexMask1D, b"mask"), 1)  # mask is texture unit 1

        # set pointers to visible textures
        GL.glClientActiveTexture(GL.GL_TEXTURE0)
        arrPointer = self._visibleTexture.ctypes.data_as(
            ctypes.POINTER(ctypes.c_float))
        GL.glTexCoordPointer(2, GL.GL_FLOAT, 0, arrPointer)
        GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        # then bind main texture
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texID)
        GL.glEnable(GL.GL_TEXTURE_2D)

        # mask
        GL.glClientActiveTexture(GL.GL_TEXTURE1)
        arrPointer = self._visibleMask.ctypes.data_as(
            ctypes.POINTER(ctypes.c_float))
        GL.glTexCoordPointer(1, GL.GL_FLOAT, 0, arrPointer)
        GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        # and mask
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_1D, self._maskID)
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glEnable(GL.GL_TEXTURE_1D)

        # do the drawing
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glDrawArrays(GL.GL_TRIANGLES, 0, self._nVisible * 3)
        # disable set states
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        GL.glDisable(GL.GL_TEXTURE_2D)

        GL.glUseProgram(0)
        # setup the shaderprogram
        GL.glEndList()

    def _updateListNoShaders(self):
        """The user shouldn't need this method since it gets called
        after every call to .set() Basically it updates the OpenGL
        representation of your stimulus if some parameter of the
        stimulus changes. Call it if you change a property manually
        rather than using the .set() command
        """
        self._needUpdate = False
        GL.glNewList(self._listID, GL.GL_COMPILE)
        # glColor can interfere with multitextures
        GL.glColor4f(1.0, 1.0, 1.0, self.opacity)

        # assign vertex array
        GL.glVertexPointer(2, GL.GL_DOUBLE, 0, self.verticesPix.ctypes)
        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)

        # bind and enable textures
        # main texture
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._texID)
        GL.glEnable(GL.GL_TEXTURE_2D)
        # mask
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glBindTexture(GL.GL_TEXTURE_1D, self._maskID)
        GL.glDisable(GL.GL_TEXTURE_2D)
        GL.glEnable(GL.GL_TEXTURE_1D)

        # set pointers to visible textures
        # mask
        GL.glClientActiveTexture(GL.GL_TEXTURE1)
        GL.glTexCoordPointer(2, GL.GL_DOUBLE, 0, self._visibleMask.ctypes)
        GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        # texture
        GL.glClientActiveTexture(GL.GL_TEXTURE0)
        GL.glTexCoordPointer(2, GL.GL_DOUBLE, 0, self._visibleTexture.ctypes)
        GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)

        # do the drawing
        GL.glDrawArrays(GL.GL_TRIANGLES, 0, self._nVisible)

        # disable set states
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY)

        GL.glEndList()

    def __del__(self):
        """Remove textures from graphics card to prevent crash
        """
        try:
            if not self.useShaders:
                GL.glDeleteLists(self._listID, 1)
            self.clearTextures()
        except (ImportError, ModuleNotFoundError, TypeError):
            pass  # has probably been garbage-collected already



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from builtins import object
from psychopy import logging, exceptions
from psychopy.constants import (STARTED, PLAYING, PAUSED, FINISHED, STOPPED,
                                NOT_STARTED, FOREVER)
from psychopy.tools import attributetools
from ._base import _SoundBase

try:
    import pysoundcard as soundcard
    import soundfile as sndfile
except ImportError as err:
    # convert this import error to our own, pysoundcard probably not installed
    raise exceptions.DependencyError(repr(err.msg))

import numpy
from os import path
import weakref


def init(rate=44100, stereo=True, buffer=128):
    pass
    # for compatibility with other backends but not needed

def getDevices(kind=None):
    """Returns a dict of dict of audio devices of specified `kind`

    The dict keys are names and items are dicts of properties
    """
    devs = {}
    for ii, dev in enumerate(soundcard.device_info()):
        if (dev['max_output_channels']==0 and kind=='output' or
                dev['max_input_channels']==0 and kind=='input'):
            continue
        # newline characters must be removed
        devName = dev['name'].replace('\r\n','')
        devs[devName] = dev
        dev['id'] = ii
    return devs

# these will be controlled by sound.__init__.py
defaultInput = None
defaultOutput = None


class _PySoundCallbackClass(object):
    """To use callbacks without creating circular references we need a
    callback class.

    Both the Stream and the sound object (SoundPySoundCard) point to this.

    This receives data and current sample from SoundPySoundCard and is
    stored by the c functions in the pysoundcard.Stream. We can't store
    a reference here to the original sound instance that created it (or
    we would create a circular ref again).

    """

    def __init__(self, sndInstance):
        self.status = NOT_STARTED
        self._sampleIndex = 0
        self.bufferSize = sndInstance.bufferSize
        self.sndInstance = weakref.ref(sndInstance)

    def fillBuffer(self, inData, outData, timeInfo, status):
        # inData to record from a buffer(?)
        # outData a buffer to write to (length of self.bufferSize)
        # timeInfo is a dict
        # status = 0 unless

        # In tests on a macbook air this function takes around 7microsec
        # to run so shouldn't impact on performance. Tested with this code:
        #    s1 = sound.SoundPySoundCard(secs=10, bufferSize=bSize)
        #    s1.play()
        #
        #    inDat = numpy.zeros([bSize,2], dtype='f')
        #    outDat = numpy.zeros([bSize,2], dtype='f')
        #    t0 = time.time()
        #    for n in range(nRuns):
        #        s1._callbacks.fillBuffer(inDat, outDat,
        #               time_info=None, status=0)
        #    print("%fms per repeat" %((time.time()-t0)*1000/nRuns))
        snd = self.sndInstance()
        chansIn, chansOut = snd._stream.channels
        nSamples = len(snd.sndArr)
        if snd.status == STOPPED:
            outData[:] = 0
            return soundcard.abort_flag
        if self._sampleIndex + self.bufferSize > nSamples:
            outData[:] = 0  # set buffer to zero
            # then insert the data
            place = nSamples - self._sampleIndex
            outData[0:place, :] = snd.sndArr[self._sampleIndex:, :]
            self._sampleIndex = nSamples
            snd._onEOS()
            return soundcard.abort_flag
        else:
            place = self._sampleIndex + self.bufferSize
            outData[:, :] = snd.sndArr[self._sampleIndex:place, :] * snd.volume
            self._sampleIndex += self.bufferSize
            return soundcard.continue_flag

    def eos(self, log=True):
        """This is potentially given directly to the paStream but we don't use
        it. Instead we're calling our own Sound.eos() from within the
        fillBuffer callback
        """
        self.sndInstance()._onEOS()


class SoundPySoundCard(_SoundBase):

    def __init__(self, value="C", secs=0.5, octave=4, sampleRate=44100,
                 bits=None, name='', autoLog=True, loops=0, bufferSize=128,
                 volume=1, stereo=True):
        """Create a sound and get ready to play

        :parameters:

            value: can be a number, string or an array:
                * If it's a number then a tone will be generated at that
                  frequency in Hz.
                * It could be a string for a note ('A', 'Bfl', 'B', 'C',
                  'Csh', ...). Then you may want to specify which octave
                * Or a string could represent a filename in the current
                location, or mediaLocation, or a full path combo
                * Or by giving an Nx2 numpy array of floats (-1:1) you can
                specify the sound yourself as a waveform

            secs: duration (only relevant if the value is a note name
                  or a frequency value)

            octave: is only relevant if the value is a note name.
                Middle octave of a piano is 4. Most computers won't
                output sounds in the bottom octave (1) and the top
                octave (8) is generally painful

            sampleRate: int (default = 44100)
                Will be ignored if a file is used for the sound and the
                sampleRate can be determined from the file

            name: string
                Only used for logging purposes

            autoLog: bool (default = true)
                Determines whether changes to the object should be logged
                by default

            loops: int (default = 0)
                number of times to loop (0 means play once, -1 means loop
                forever)

            bufferSize: int (default = 128)
                How many samples should be loaded at a time to the sound
                buffer. A larger number will reduce speed to play a sound.
                If too small then audio artifacts will be heard where the
                buffer ran empty

            bits:
                currently serves no purpose (exists for backwards
                compatibility)

            volume: 0-1.0

            stereo:
                currently serves no purpose (exists for backwards
                compatibility)
        """
        self.name = name  # only needed for autoLogging
        self.autoLog = autoLog

        self.sampleRate = sampleRate
        self.bufferSize = bufferSize
        self.volume = volume

        # try to create sound
        self._snd = None
        # distinguish the loops requested from loops actual because of
        # infinite tones (which have many loops but none requested)
        # -1 for infinite or a number of loops
        self.requestedLoops = self.loops = int(loops)
        self.setSound(value=value, secs=secs, octave=octave)

    def play(self, fromStart=True, log=True, loops=None, when=None):
        """Starts playing the sound on an available channel.

        :Parameters:

            fromStart : bool
                Not yet implemented.
            log : bool
                Whether or not to log the playback event.
            loops : int
                How many times to repeat the sound after it plays once. If
                `loops` == -1, the sound will repeat indefinitely until
                stopped.
            when: not used
                Included for compatibility purposes

        :Notes:

            If no sound channels are available, it will not play and return
            None. This runs off a separate thread, i.e. your code won't
            wait for the sound to finish before continuing. You need to
            use a psychopy.core.wait() command if you want things to pause.
            If you call play() whiles something is already playing the sounds
            will be played over each other.

        """
        if loops is not None:
            self.loops = loops
        self._stream.start()
        self.status = STARTED
        if log and self.autoLog:
            logging.exp("Sound %s started" % (self.name), obj=self)
        return self

    def stop(self, log=True):
        """Stops the sound immediately"""
        self._stream.abort()  # _stream.stop() finishes current buffer
        self.status = STOPPED
        if log and self.autoLog:
            logging.exp("Sound %s stopped" % (self.name), obj=self)

    def fadeOut(self, mSecs):
        """fades out the sound (when playing) over mSecs.
        Don't know why you would do this in psychophysics but it's easy
        and fun to include as a possibility :)
        """
        pass  # todo
        self.status = STOPPED

    def getDuration(self):
        """Get's the duration of the current sound in secs
        """
        pass  # todo

    @attributetools.attributeSetter
    def volume(self, volume):
        """Returns the current volume of the sound (0.0:1.0)
        """
        self.__dict__['volume'] = volume

    def setVolume(self, value, operation="", log=None):
        """Sets the current volume of the sound (0.0:1.0)
        """
        attributetools.setAttribute(self, 'volume', value, log, operation)
        return value  # this is returned for historical reasons

    def _setSndFromFile(self, fileName):
        # load the file
        if not path.isfile(fileName):
            msg = "Sound file %s could not be found." % fileName
            logging.error(msg)
            raise ValueError(msg)
        self.fileName = fileName
        # in case a tone with inf loops had been used before
        self.loops = self.requestedLoops
        try:
            self.sndFile = sndfile.SoundFile(fileName)
            sndArr = self.sndFile.read()
            self.sndFile.close()
            self._setSndFromArray(sndArr)

        except Exception:
            msg = "Sound file %s could not be opened using pysoundcard for sound."
            logging.error(msg % fileName)
            raise ValueError(msg % fileName)

    @property
    def status(self):
        # NB this is stored by the _callbacks class for fast access when
        # data buffer needs filling (_callbacks class does not have a
        # reference back here)
        return self.__dict__['status']

    @status.setter
    def status(self, status):
        self.__dict__['status'] = status

    def _setSndFromArray(self, thisArray):
        """For pysoundcard all sounds are ultimately played as an array so
        other setSound methods are going to call this having created an arr
        """
        self._callbacks = _PySoundCallbackClass(sndInstance=self)
        if defaultOutput is not None and type(defaultOutput) != int:
            devs = getDevices()
            if defaultOutput not in devs:
                raise ValueError("Attempted to set use device {!r} to "
                                 "a device that is not available".format(defaultOutput))
            else:
                device = devs[defaultOutput]['id']
        else:
            device = defaultOutput
        self._stream = soundcard.Stream(samplerate=self.sampleRate,
                                        device=device,
                                        blocksize=self.bufferSize,
                                        channels=1,
                                        callback=self._callbacks.fillBuffer)
        self._snd = self._stream
        chansIn, chansOut = self._stream.channels
        if chansOut > 1 and thisArray.ndim == 1:
            # make mono sound stereo
            self.sndArr = numpy.resize(thisArray, [2, len(thisArray)]).T
        else:
            self.sndArr = numpy.asarray(thisArray)
        self._nSamples = thisArray.shape[0]
        # set to run from the start:
        self.seek(0)

    def seek(self, t):
        self._sampleIndex = t * self.sampleRate

    def _onEOS(self, log=True):
        if log and self.autoLog:
            logging.exp("Sound %s finished" % (self.name), obj=self)
        self.status = FINISHED

    def __del__(self):
        self._stream.close()
